# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Defines functionality for the evaluation of predictions."""

import json
import logging
import pathlib
import typing

import credible.curves
import credible.plot
import h5py
import numpy
import numpy.typing
import tabulate
from tqdm import tqdm

logger = logging.getLogger(__name__)

SUPPORTED_METRIC_TYPE = typing.Literal[
    "precision", "recall", "specificity", "accuracy", "jaccard", "f1"
]
"""Supported metrics for evaluation of counts."""


def _tricky_division(n: float, d: float):
    """Divide n by d, or 0.0 in case of a division by zero.

    Parameters
    ----------
    n
        The number to divide.
    d
        The divisor.

    Returns
    -------
        Result of the division.
    """

    return n / (d + (d == 0))


def precision(tp: int, fp: int, tn: int, fn: int) -> float:
    """Calculate the precision given true/false positive/negative counts.

    P, AKA positive predictive value (PPV).  It corresponds arithmetically to
    ``tp/(tp+fp)``.  In the case ``tp+fp == 0``, this function returns zero for
    precision.

    Parameters
    ----------
    tp
        True positive count, AKA "hit".
    fp
        False positive count, AKA "false alarm", or "Type I error".
    tn
        True negative count, AKA "correct rejection".
    fn
        False Negative count, AKA "miss", or "Type II error".

    Returns
    -------
        The precision.
    """
    return _tricky_division(tp, tp + fp)


def recall(tp: int, fp: int, tn: int, fn: int) -> float:
    """Calculate the recall given true/false positive/negative counts.

    R, AKA sensitivity, hit rate, or true positive rate (TPR).  It corresponds
    arithmetically to ``tp/(tp+fn)``.  In the special case where ``tp+fn ==
    0``, this function returns zero for recall.

    Parameters
    ----------
    tp
        True positive count, AKA "hit".
    fp
        False positive count, AKA "false alarm", or "Type I error".
    tn
        True negative count, AKA "correct rejection".
    fn
        False Negative count, AKA "miss", or "Type II error".

    Returns
    -------
        The recall.
    """
    return _tricky_division(tp, tp + fn)


def specificity(tp: int, fp: int, tn: int, fn: int) -> float:
    """Calculate the specificity given true/false positive/negative counts.

    S, AKA selectivity or true negative rate (TNR).  It corresponds
    arithmetically to ``tn/(tn+fp)``.  In the special case where ``tn+fp ==
    0``, this function returns zero for specificity.

    Parameters
    ----------
    tp
        True positive count, AKA "hit".
    fp
        False positive count, AKA "false alarm", or "Type I error".
    tn
        True negative count, AKA "correct rejection".
    fn
        False Negative count, AKA "miss", or "Type II error".

    Returns
    -------
        The specificity.
    """
    return _tricky_division(tn, fp + tn)


def accuracy(tp: int, fp: int, tn: int, fn: int) -> float:
    """Calculate the accuracy given true/false positive/negative counts.

    A, see `Accuracy
    <https://en.wikipedia.org/wiki/Evaluation_of_binary_classifiers>`_. is the
    proportion of correct predictions (both true positives and true negatives)
    among the total number of pixels examined.  It corresponds arithmetically
    to ``(tp+tn)/(tp+tn+fp+fn)``.  This measure includes both true-negatives
    and positives in the numerator, what makes it sensitive to data or regions
    without annotations.

    Parameters
    ----------
    tp
        True positive count, AKA "hit".
    fp
        False positive count, AKA "false alarm", or "Type I error".
    tn
        True negative count, AKA "correct rejection".
    fn
        False Negative count, AKA "miss", or "Type II error".

    Returns
    -------
        The accuracy.
    """
    return _tricky_division(tp + tn, tp + fp + fn + tn)


def jaccard(tp: int, fp: int, tn: int, fn: int) -> float:
    """Calculate the Jaccard index given true/false positive/negative counts.

    J, see `Jaccard Index or Similarity
    <https://en.wikipedia.org/wiki/Jaccard_index>`_.  It corresponds
    arithmetically to ``tp/(tp+fp+fn)``.  In the special case where ``tn+fp+fn
    == 0``, this function returns zero for the Jaccard index. The Jaccard index
    depends on a TP-only numerator, similarly to the F1 score.  For regions
    where there are no annotations, the Jaccard index will always be zero,
    irrespective of the model output.  Accuracy may be a better proxy if one
    needs to consider the true abscence of annotations in a region as part of
    the measure.

    Parameters
    ----------
    tp
        True positive count, AKA "hit".
    fp
        False positive count, AKA "false alarm", or "Type I error".
    tn
        True negative count, AKA "correct rejection".
    fn
        False Negative count, AKA "miss", or "Type II error".

    Returns
    -------
        The Jaccard index.
    """
    return _tricky_division(tp, tp + fp + fn)


def f1_score(tp: int, fp: int, tn: int, fn: int) -> float:
    """Calculate the F1 score given true/false positive/negative counts.

    F1, see `F1-score <https://en.wikipedia.org/wiki/F1_score>`_.  It
    corresponds arithmetically to ``2*P*R/(P+R)`` or ``2*tp/(2*tp+fp+fn)``. In
    the special case where ``P+R == (2*tp+fp+fn) == 0``, this function returns
    zero for the Jaccard index.  The F1 or Dice score depends on a TP-only
    numerator, similarly to the Jaccard index.  For regions where there are no
    annotations, the F1-score will always be zero, irrespective of the model
    output.  Accuracy may be a better proxy if one needs to consider the true
    abscence of annotations in a region as part of the measure.

    Parameters
    ----------
    tp
        True positive count, AKA "hit".
    fp
        False positive count, AKA "false alarm", or "Type I error".
    tn
        True negative count, AKA "correct rejection".
    fn
        False Negative count, AKA "miss", or "Type II error".

    Returns
    -------
        The F1-score.
    """
    return _tricky_division(2 * tp, (2 * tp) + fp + fn)


_METRIC_MAPPING = {
    k: v
    for k, v in zip(
        typing.get_args(SUPPORTED_METRIC_TYPE),
        [precision, recall, specificity, accuracy, jaccard, f1_score],
    )
}
"""Maps supported metric names and callables implementing them."""


def name2metric(
    name: SUPPORTED_METRIC_TYPE,
) -> typing.Callable[[int, int, int, int], float]:
    """Convert a string name to a callable for summarizing counts.

    Parameters
    ----------
    name
        The name of the metric to be looked up.

    Returns
    -------
        A callable that summarizes counts into single floating-point number.
    """
    return _METRIC_MAPPING[name]


def all_metrics(tp: int, fp: int, tn: int, fn: int) -> list[float]:
    """Compute all available metrics at once.

    Parameters
    ----------
    tp
        True positive count, AKA "hit".
    fp
        False positive count, AKA "false alarm", or "Type I error".
    tn
        True negative count, AKA "correct rejection".
    fn
        False Negative count, AKA "miss", or "Type II error".

    Returns
    -------
        All supported metrics in the order defined by
        :py:obj:`SUPPORTED_METRIC_TYPE`.
    """
    return [k(tp, fp, tn, fn) for k in _METRIC_MAPPING.values()]


def tfpn_masks(
    pred: numpy.typing.NDArray[numpy.float32],
    gt: numpy.typing.NDArray[numpy.bool_],
    threshold: float,
) -> tuple[
    numpy.typing.NDArray[numpy.bool_],
    numpy.typing.NDArray[numpy.bool_],
    numpy.typing.NDArray[numpy.bool_],
    numpy.typing.NDArray[numpy.bool_],
]:
    """Calculate true and false positives and negatives.

    All input arrays should have matching sizes.

    Parameters
    ----------
    pred
        Pixel-wise predictions as output by your model.
    gt
        Ground-truth (annotations).
    threshold
        A particular threshold in which to calculate the performance
        measures.  Values at this threshold are counted as positives.

    Returns
    -------
    tp
        Boolean array with true positives, considering all observations.
    fp
        Boolean array with false positives, considering all observations.
    tn
        Boolean array with true negatives, considering all observations.
    fn
        Boolean array with false negatives, considering all observations.
    """

    binary_pred = pred >= threshold
    tp = numpy.logical_and(binary_pred, gt)
    fp = numpy.logical_and(binary_pred, numpy.logical_not(gt))
    tn = numpy.logical_and(numpy.logical_not(binary_pred), numpy.logical_not(gt))
    fn = numpy.logical_and(numpy.logical_not(binary_pred), gt)
    return tp, fp, tn, fn


def get_counts_for_threshold(
    pred: numpy.typing.NDArray[numpy.float32],
    gt: numpy.typing.NDArray[numpy.bool_],
    mask: numpy.typing.NDArray[numpy.bool_],
    threshold: float,
) -> tuple[int, int, int, int]:
    """Calculate counts on one single sample, for a specific threshold.

    Parameters
    ----------
    pred
        Array with pixel-wise predictions.
    gt
        Array with ground-truth (annotations).
    mask
        Array with region mask marking parts to ignore.
    threshold
        A particular threshold in which to calculate the performance
        measures.

    Returns
    -------
        The true positives, false positives, true negatives and false
        negatives, in that order.
    """

    tp, fp, tn, fn = tfpn_masks(pred, gt, threshold)
    tp = numpy.logical_and(tp, mask)
    fp = numpy.logical_and(fp, mask)
    tn = numpy.logical_and(tn, mask)
    fn = numpy.logical_and(fn, mask)
    return tp.sum(), fp.sum(), tn.sum(), fn.sum()


def load_count(
    prediction_path: pathlib.Path,
    predictions: typing.Sequence[str],
    thresholds: numpy.typing.NDArray[numpy.float64],
) -> numpy.typing.NDArray[numpy.uint64]:
    """Count true/false positive/negatives for the subset.

    This function will load predictions from their store location and will
    **cumulatively** count the number of true positives, false positives, true
    negatives and false negatives across the various ``thresholds``.  This
    alternative provides a memory-bound way to compute the performance of
    splits with potentially very large images or including a large/very large
    number of samples.  Unfortunately, sklearn does not provide functions to
    compute standard metrics from true/false positive/negative counts, which
    implies one needs to make use of further functions defined in this module
    to compute such metrics.  Alternatively, you may look into
    :py:func:`load_predictions`, if you want to use sklearn functions to
    compute metrics.

    Parameters
    ----------
    prediction_path
        Base directory where the prediction files (HDF5) were stored.
    predictions
        A list of relative sample prediction paths to consider for measurement.
    thresholds
        A sequence of thresholds to be applied on ``predictions``, when
        evaluating true/false positive/negative counts.

    Returns
    -------
        A 2-D array with shape ``(len(thresholds), 4)``, where each row
        contains to the counts of true positives, false positives, true
        negatives and false negatives, for the related threshold, and for the
        **whole dataset**.
    """
    data = numpy.zeros((len(thresholds), 4), dtype=numpy.uint64)
    for sample in tqdm(predictions, desc="sample"):
        with h5py.File(prediction_path / sample[1], "r") as f:
            pred = numpy.array(f["prediction"])  # float32
            gt = numpy.array(f["target"])  # boolean
            mask = numpy.array(f["mask"])  # boolean
        data += numpy.array(
            [get_counts_for_threshold(pred, gt, mask, k) for k in thresholds],
            dtype=numpy.uint64,
        )
    return data


def load_predictions(
    prediction_path: pathlib.Path,
    predictions: typing.Sequence[str],
) -> tuple[numpy.typing.NDArray[numpy.float32], numpy.typing.NDArray[numpy.bool_]]:
    """Load predictions and ground-truth from HDF5 files.

    Loading pixel-data as simple binary predictions with associated labels
    allows using sklearn library to compute most metrics defined in this
    module.  Note however that computing metrics this way requires
    pre-allocation of a potentially large vector, which depends on the number
    of samples and the size of said samples.  This may not work well for very
    large datasets of large/very large images.  Currently, the evaluation
    system uses :py:func:`load_count` instead, which loads and pre-computes the
    number of true/false positives/negatives using a list of candidate
    thresholds.

    Parameters
    ----------
    prediction_path
        Base directory where the prediction files (HDF5) were stored.
    predictions
        A list of relative sample prediction paths to consider for measurement.

    Returns
    -------
        Two 1-D arrays containing a linearized version of pixel predictions
        (probability) and matching ground-truth.
    """

    # peak prediction size and number of samples
    with h5py.File(prediction_path / predictions[0][1], "r") as f:
        data: h5py.Dataset = typing.cast(h5py.Dataset, f["prediction"])
        elements = numpy.array(data.shape).prod()
    size = len(predictions) * elements
    logger.info(
        f"Data loading will require ({elements} x {len(predictions)} x 5 =) "
        f"{size*5/(1024*1024):.0f} MB of RAM"
    )

    # now load the data
    pred_array = numpy.empty((size,), dtype=numpy.float32)
    gt_array = numpy.empty((size,), dtype=numpy.bool_)
    for i, sample in enumerate(tqdm(predictions, desc="sample")):
        with h5py.File(prediction_path / sample[1], "r") as f:
            mask = numpy.array(f["mask"])  # boolean
            pred = numpy.array(f["prediction"])  # float32
            pred *= mask.astype(numpy.float32)
            gt = numpy.array(f["target"])  # boolean
            gt &= mask
            pred_array[i * elements : (i + 1) * elements] = pred.flatten()
            gt_array[i * elements : (i + 1) * elements] = gt.flatten()

    return pred_array, gt_array


def compute_metric(
    counts: numpy.typing.NDArray[numpy.uint64],
    metric: typing.Callable[[int, int, int, int], float]
    | typing.Callable[[int, int, int, int], tuple[float, ...]],
) -> numpy.typing.NDArray[numpy.float64]:
    """Compute ``metric`` for every row of ``counts``.

    Parameters
    ----------
    counts
        A 2-D array with shape ``(*, 4)``, where each row contains to the
        counts of true positives, false positives, true negatives and false
        negatives, that need to be evaluated.
    metric
        A callable that takes 4 integers representing true positives, false
        positives, true negatives and false negatives, and outputs one or more
        floating-point metrics.

    Returns
    -------
        An 1-D array containing the provided metric computed alongside the
        first dimension and as many columns as ```metric`` provides in each
        call.
    """
    return numpy.array([metric(*k) for k in counts], dtype=numpy.float64)


def validate_threshold(threshold: float | str, splits: list[str]):
    """Validate the user threshold selection and returns parsed threshold.

    Parameters
    ----------
    threshold
        The threshold to validate.
    splits
        List of available splits.

    Returns
    -------
        The validated threshold.
    """
    try:
        # we try to convert it to float first
        threshold = float(threshold)
        if threshold < 0.0 or threshold > 1.0:
            raise ValueError("Float thresholds must be within range [0.0, 1.0]")
    except ValueError:
        if threshold not in splits:
            raise ValueError(
                f"Text thresholds should match dataset names, "
                f"but {threshold} is not available among the datasets provided ("
                f"({', '.join(splits)})"
            )

    return threshold


def compare_annotators(
    a1: pathlib.Path, a2: pathlib.Path
) -> dict[str, dict[str, float]]:
    """Compare annotators and outputs all supported metrics.

    Parameters
    ----------
    a1
        Annotator 1 annotations in the form of a JSON file mapping split-names
        ot list of lists, each containing the sample name and the (relative)
        location of an HDF5 file containing at least one boolean dataset named
        ``target``.  This dataset is considered as the annotations from the
        first annotator.  If a boolean ``mask`` is available, it is also
        loaded.  All elements outside the mask are not considered during the
        metrics calculations.
    a2
        Annotator 1 annotations in the form of a JSON file mapping split-names
        ot list of lists, each containing the sample name and the (relative)
        location of an HDF5 file containing at least one boolean dataset named
        ``target``.  This dataset is considered as the annotations from the
        second annotator.

    Returns
    -------
        A dictionary that maps split-names to another dictionary with metric
        names and values computed by comparing the targets in ``a1`` and
        ``a2``.
    """

    retval: dict[str, dict[str, float]] = {}

    with a1.open("r") as f:
        a1_data = json.load(f)

    with a2.open("r") as f:
        a2_data = json.load(f)

    metrics_available = list(typing.get_args(SUPPORTED_METRIC_TYPE))

    for split_name, samples in a1_data.items():
        if split_name not in a2_data:
            continue

        for _, hdf5_path in samples:
            with h5py.File(a1.parent / hdf5_path, "r") as f:
                t1 = numpy.array(f["target"])
                mask = numpy.array(f["mask"])
            with h5py.File(a2.parent / hdf5_path, "r") as f:
                t2 = numpy.array(f["target"])

            tp, fp, tn, fn = get_counts_for_threshold(t2, t1, mask, 0.5)
            base_metrics = all_metrics(tp, fp, tn, fn)
            retval[split_name] = {k: v for k, v in zip(metrics_available, base_metrics)}

    return retval


def run(
    predictions: pathlib.Path,
    steps: int,
    threshold: str | float,
    metric: SUPPORTED_METRIC_TYPE,
) -> tuple[dict[str, dict[str, typing.Any]], float]:
    """Evaluate a segmentation model.

    Parameters
    ----------
    predictions
        Path to the file ``predictions.json``, containing the list of
        predictions to be evaluated.
    steps
        The number of steps between ``[0, 1]`` to build a threshold list
        from.  This list will be applied to the probability outputs and
        true/false positive/negative counts generated from those.
    threshold
        Which threshold to apply when generating unary summaries of the
        performance.  This can be a value between ``[0, 1]``, or the name
        of a split in ``predictions`` where a threshold will be calculated
        at.
    metric
        The name of a supported metric that will be used to evaluate the
        best threshold from a threshold-list uniformily split in ``steps``,
        and for which unary summaries are generated.

    Returns
    -------
        A JSON-able summary with all figures of merit pre-caculated, for
        all splits.  This is a dictionary where keys are split-names contained
        in ``predictions``, and values are dictionaries with the following
        keys:

            * ``counts``: dictionary where keys are thresholds, and values are
              sequence of integers containing the TP, FP, TN, FN (in this order).

            * ``roc_auc``: a float indicating the area under the ROC curve
              for the split.  It is calculated using a trapezoidal rule.

            * ``average_precision``: a float indicating the area under the
              precision-recall curve, calculated using a rectangle rule.

            * ``curves``: dictionary with 2 keys:

              * ``roc``: dictionary with 3 keys:

                * ``fpr``: a list of floats with the false-positive rate
                * ``tpr``: a list of floats with the true-positive rate
                * ``thresholds``: a list of thresholds uniformily separated by
                  ``steps``, at which both ``fpr`` and ``tpr`` are evaluated.
              * ``precision_recall``: a dictionary with 3 keys:

                * ``precision``: a list of floats with the precision
                * ``recall``: a list of floats with the recall
                * ``thresholds``: a list of thresholds uniformily separated by
                  ``steps``, at which both ``precision`` and ``recall`` are
                  evaluated.

            * ``threshold_a_priori``: boolean indicating if the threshold for unary
              metrics where computed with a threshold chosen a priori or a
              posteriori in this split.

            * ``<metric-name>``: a float representing the supported metric at the
              threshold that maximizes ``metric``.  There will be one entry of this
              type for each of the :py:obj:`SUPPORTED_METRIC_TYPE`'s.

        Also returns the threshold considered for all splits.
    """

    with predictions.open("r") as f:
        predict_data = json.load(f)

    threshold = validate_threshold(threshold, predict_data)
    threshold_list = numpy.arange(
        0.0, (1.0 + 1 / steps), 1 / steps, dtype=numpy.float64
    )

    # Holds all computed data.  Format <split-name: str> -> <split-data: dict>
    eval_json_data: dict[str, dict[str, typing.Any]] = {}

    # Compute counts for various splits.
    for split_name, samples in predict_data.items():
        logger.info(
            f"Counting true/false positive/negatives at split `{split_name}`..."
        )
        counts = load_count(predictions.parent, samples, threshold_list)

        logger.info(f"Evaluating performance curves/metrics at split `{split_name}`...")
        fpr_curve = 1.0 - numpy.array([specificity(*k) for k in counts])
        recall_curve = tpr_curve = numpy.array([recall(*k) for k in counts])
        precision_curve = numpy.array([precision(*k) for k in counts])

        # correction when precision is very small
        precision_curve[
            numpy.logical_and(precision_curve < 1e-8, recall_curve < 1e-8)
        ] = 1.0

        # populates data to be recorded in JSON format
        eval_json_data.setdefault(split_name, {})["samples"] = len(samples)
        eval_json_data.setdefault(split_name, {})["counts"] = {
            k: v for k, v in zip(threshold_list, counts)
        }
        eval_json_data.setdefault(split_name, {})["roc_auc"] = (
            credible.curves.area_under_the_curve((fpr_curve, tpr_curve))
        )
        eval_json_data.setdefault(split_name, {})["average_precision"] = (
            credible.curves.average_metric((precision_curve, recall_curve))
        )
        eval_json_data.setdefault(split_name, {})["curves"] = dict(
            roc=dict(fpr=fpr_curve, tpr=tpr_curve, thresholds=threshold_list),
            precision_recall=dict(
                precision=precision_curve,
                recall=recall_curve,
                thresholds=threshold_list,
            ),
        )

    # Computes argmax in the designated split "counts" (typically "validation"),
    # where the chosen metric reaches its **maximum**.
    if isinstance(threshold, str):
        # Compute threshold on specified split, if required
        logger.info(f"Evaluating threshold on split `{threshold}` using " f"`{metric}`")
        metric_list = compute_metric(
            eval_json_data[threshold]["counts"].values(),
            name2metric(typing.cast(SUPPORTED_METRIC_TYPE, metric)),
        )
        threshold_index = metric_list.argmax()

        # Reset list of how thresholds are calculated on the recorded split
        for split_name in predict_data.keys():
            if split_name == threshold:
                eval_json_data[split_name]["threshold_a_priori"] = False
            else:
                eval_json_data[split_name]["threshold_a_posteriori"] = True

    else:
        # must figure out the closest threshold from the list we are using
        threshold_index = (numpy.abs(threshold_list - threshold)).argmin()

        # Reset list of how thresholds are calculated on the recorded split
        for split_name in predict_data.keys():
            eval_json_data[split_name]["threshold_a_priori"] = True

    logger.info(f"Set --threshold={threshold_list[threshold_index]:.4f}")

    # Computes all available metrics on the designated threshold, across all
    # splits
    # Populates <split-name: str> -> <metric-name: SUPPORTED_METRIC_TYPE> ->
    # float
    metrics_available = list(typing.get_args(SUPPORTED_METRIC_TYPE))
    for split_name in predict_data.keys():
        logger.info(
            f"Computing metrics on split `{split_name}` at "
            f"threshold={threshold_list[threshold_index]:.4f}..."
        )
        base_metrics = all_metrics(
            *(list(eval_json_data[split_name]["counts"].values())[threshold_index])
        )
        eval_json_data[split_name]["threshold"] = threshold_list[threshold_index]
        eval_json_data[split_name].update(
            {k: v for k, v in zip(metrics_available, base_metrics)}
        )

    return eval_json_data, threshold_list[threshold_index]


def make_table(
    eval_data: dict[str, dict[str, typing.Any]], threshold: float, format_: str
) -> str:
    """Extract and format table from pre-computed evaluation data.

    Extracts elements from ``eval_data`` that can be displayed on a
    terminal-style table, format, and returns it.

    Parameters
    ----------
    eval_data
        Evaluation data as returned by :py:func:`run`.
    threshold
        The threshold value used to compute unary metrics on all splits.
    format_
        A supported tabulate format.

    Returns
    -------
        A string representation of a table.
    """

    # Extracts elements from ``eval_json_data`` that can be displayed on a
    # terminal-style table, format, and print to screen.  Record the table into
    # a file for later usage.
    metrics_available = list(typing.get_args(SUPPORTED_METRIC_TYPE))
    table_headers = (
        ["subset", "samples", "threshold"] + metrics_available + ["auc_roc", "avg.prec"]
    )

    table_data = []
    for split_name, data in eval_data.items():
        base_metrics = [data[k] for k in metrics_available]
        table_data.append(
            [split_name, data["samples"], threshold]
            + base_metrics
            + [data["roc_auc"], data["average_precision"]]
        )

        if "second_annotator" in data:
            table_data.append(
                [split_name, "annotator/2"]
                + list(data["second_annotator"].values())
                + [None, None]
            )

    return tabulate.tabulate(
        table_data,
        table_headers,
        tablefmt=format_,
        floatfmt=".3f",
        stralign="right",
    )


def make_plots(eval_data: dict[str, dict[str, typing.Any]]) -> list:
    """Create plots for all curves in ``eval_data``.

    Parameters
    ----------
    eval_data
        Evaluation data as returned by :py:func:`run`.

    Returns
    -------
        A list of figures to record to file
    """
    retval = []

    with credible.plot.tight_layout(
        ("False Positive Rate", "True Positive Rate"), "ROC"
    ) as (fig, ax):
        for split_name, data in eval_data.items():
            ax.plot(
                data["curves"]["roc"]["fpr"],
                data["curves"]["roc"]["tpr"],
                label=f"{split_name} (AUC: {data['roc_auc']:.2f})",
            )

            if "second_annotator" in data:
                fpr = 1 - data["second_annotator"]["specificity"]
                tpr = data["second_annotator"]["recall"]
                ax.plot(
                    fpr,
                    tpr,
                    linestyle="none",
                    marker="*",
                    markersize=8,
                    label=f"{split_name} (annotator/2)",
                )

            ax.legend(loc="best", fancybox=True, framealpha=0.7)
        retval.append(fig)

    with credible.plot.tight_layout_f1iso(
        ("Recall", "Precision"), "Precison-Recall"
    ) as (fig, ax):
        for split_name, data in eval_data.items():
            ax.plot(
                data["curves"]["precision_recall"]["recall"],
                data["curves"]["precision_recall"]["precision"],
                label=f"{split_name} (AP: {data['average_precision']:.2f})",
            )

            if "second_annotator" in data:
                recall = data["second_annotator"]["recall"]
                precision = data["second_annotator"]["precision"]
                ax.plot(
                    recall,
                    precision,
                    linestyle="none",
                    marker="*",
                    markersize=8,
                    label=f"{split_name} (annotator/2)",
                )

            ax.legend(loc="best", fancybox=True, framealpha=0.7)
        retval.append(fig)

    return retval



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Extension of ``lightning.LightningDataModule`` with dictionary split loading, mini-batching, parallelisation and caching."""

import functools
import itertools
import logging
import sys
import typing

import lightning
import torch
import torch.backends
import torch.utils.data
import torchvision.transforms
import torchvision.tv_tensors
import tqdm

from .typing import (
    ConcatDatabaseSplit,
    DatabaseSplit,
    DataLoader,
    Dataset,
    RawDataLoader,
    Sample,
    TransformSequence,
)

logger = logging.getLogger(__name__)


def _sample_size_bytes(dataset: Dataset):
    """Recurse into the first sample of a dataset and figures out its total occupance in bytes.

    Parameters
    ----------
    dataset
        The dataset containing the samples to load.
    """

    def _tensor_size_bytes(t: torch.Tensor, n: str) -> int:
        """Return a tensor size in bytes.

        Parameters
        ----------
        t
            A torch Tensor.
        n
            Name of the object.

        Returns
        -------
        int
            The size of the Tensor in bytes.
        """

        logger.info(f"`{n}`: {list(t.shape)}@{t.dtype}")
        return int(t.element_size() * t.shape.numel())

    def _dict_size_bytes(d):
        """Return a dictionary size in bytes.

        Parameters
        ----------
        d
            A dictionary.

        Returns
        -------
        int
            The size of the dictionary in bytes.
        """

        size = 0
        for k, v in d.items():
            if isinstance(v, torch.Tensor):
                size += _tensor_size_bytes(v, k)

        return size

    first_sample = dataset[0]
    size = sys.getsizeof(first_sample)  # measures size of all pythonic objects
    size += _dict_size_bytes(first_sample)  # adds torch tensor sizes

    sample_size_mb = size / (1024.0 * 1024.0)
    logger.info(f"Estimated sample size: {sample_size_mb:.1f} Mb")


def _apply_iff_tv_tensor(
    data: typing.Any, transform: typing.Callable[[torch.Tensor], torch.Tensor]
):
    """Apply model_transform iff input object is TVTensor.

    Parameters
    ----------
    data
        Sample data to which apply the provided transform.
    transform
        Callable containing a single or a composition of transforms to
        potentially apply to ``data``.

    Returns
    -------
        The transformed version of ``data`` iff applicable.
    """
    if isinstance(data, torchvision.tv_tensors.TVTensor):
        return transform(data)
    return data


class _DelayedLoadingDataset(Dataset):
    """A list that loads its samples on demand.

    This list mimics a pytorch Dataset, except that raw data loading is done
    on-the-fly, as the samples are requested through the bracket operator.

    Parameters
    ----------
    raw_dataset
        An iterable containing the raw dataset samples representing one of the
        database split datasets.
    loader
        An object instance that can load samples from storage.
    transforms
        A set of transforms that should be applied on-the-fly for this dataset,
        to fit the output of the raw-data-loader to the model of interest.
    disable_pbar
        If set, disables progress bars.
    """

    def __init__(
        self,
        raw_dataset: typing.Sequence[typing.Any],
        loader: RawDataLoader,
        transforms: TransformSequence = [],
        disable_pbar: bool = False,
    ):
        self.raw_dataset = raw_dataset
        self.loader = loader
        self.transform = torchvision.transforms.Compose(transforms)
        self.disable_pbar = disable_pbar

        _sample_size_bytes(self)

    def targets(self) -> list[torch.Tensor]:
        """Return the targets for all samples in the dataset.

        Returns
        -------
            The targets for all samples in the dataset.
        """

        return [
            self.loader.target(k)
            for k in tqdm.tqdm(
                self.raw_dataset, unit="sample", disable=self.disable_pbar
            )
        ]

    def __getitem__(self, key: int) -> Sample:
        sample = self.loader.sample(self.raw_dataset[key])
        return {k: _apply_iff_tv_tensor(v, self.transform) for k, v in sample.items()}

    def __len__(self):
        return len(self.raw_dataset)

    def __iter__(self):
        for x in range(len(self)):
            yield self[x]


def _apply_loader_and_transforms(
    info: typing.Any,
    load: typing.Callable[[typing.Any], Sample],
    model_transform: typing.Callable[[torch.Tensor], torch.Tensor],
) -> Sample:
    """Local wrapper to apply raw-data loading and transformation in a single
    step.

    Parameters
    ----------
    info
        The sample information, as loaded from its raw dataset dictionary.
    load
        The raw-data loader function to use for loading the sample.
    model_transform
        A callable that will transform the loaded tensor into something
        suitable for the model it will train.  Typically, this will be a
        composed transform.

    Returns
    -------
    Sample
        The loaded and transformed sample.
    """
    sample = load(info)
    return {k: _apply_iff_tv_tensor(v, model_transform) for k, v in sample.items()}


class CachedDataset(Dataset):
    """Basically, a list of preloaded samples.

    This dataset will load all samples from the raw dataset during construction
    instead of delaying that to the indexing.  Beyond raw-data-loading,
    ``transforms`` given upon construction contribute to the cached samples.

    Parameters
    ----------
    raw_dataset
        An iterable containing the raw dataset samples representing one of the
        database split datasets.
    loader
        An object instance that can load samples and targets from storage.
    transforms
        A set of transforms that should be applied to the cached samples for
        this dataset, to fit the output of the raw-data-loader to the model of
        interest.
    parallel
        Use multiprocessing for data loading: if set to -1 (default), disables
        multiprocessing data loading.  Set to 0 to enable as many data loading
        instances as processing cores available in the system.  Set to >= 1
        to enable that many multiprocessing instances for data loading.
    multiprocessing_context
        Which implementation of the multiprocessing context to use. Options are
        defined at :py:mod:`multiprocessing`.  If set to ``None``, use the
        default for the current platform.
    disable_pbar
        If set, disables progress bars.
    """

    def __init__(
        self,
        raw_dataset: typing.Sequence[typing.Any],
        loader: RawDataLoader,
        transforms: TransformSequence = [],
        parallel: int = -1,
        multiprocessing_context: str | None = None,
        disable_pbar: bool = False,
    ):
        self.loader = functools.partial(
            _apply_loader_and_transforms,
            load=loader.sample,
            model_transform=torchvision.transforms.Compose(transforms),
        )

        if parallel >= 0 and sys.platform == "linux":
            logger.warning(
                "Parallel sample caching is currently NOT supported under Linux. "
                "Check https://gitlab.idiap.ch/medai/software/mednet/-/issues/93 "
                "for updates."
            )
            parallel = -1

        if parallel < 0:
            self.data = [
                self.loader(k)
                for k in tqdm.tqdm(raw_dataset, unit="sample", disable=disable_pbar)
            ]
        else:
            instances = parallel or torch.multiprocessing.cpu_count()
            logger.info(f"Caching dataset using {instances} processes...")
            with torch.multiprocessing.get_context(multiprocessing_context).Pool(
                instances
            ) as p:
                self.data = list(
                    tqdm.tqdm(
                        p.imap(self.loader, raw_dataset),
                        total=len(raw_dataset),
                        disable=disable_pbar,
                    ),
                )

        _sample_size_bytes(self)

    def targets(self) -> list[torch.Tensor]:
        """Return the targets for all samples in the dataset.

        Returns
        -------
            The targets for all samples in the dataset.
        """

        return [k["target"] for k in self.data]

    def __getitem__(self, key: int) -> Sample:
        return self.data[key]

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        yield from self.data


class ConcatDataset(Dataset):
    """A dataset that represents a concatenation of other cached or delayed datasets.

    Parameters
    ----------
    datasets
        An iterable over pre-instantiated datasets.
    """

    def __init__(self, datasets: typing.Sequence[Dataset]):
        self._datasets = datasets
        self._indices = [
            (i, j)  # dataset relative position, sample relative position
            for i in range(len(datasets))
            for j in range(len(datasets[i]))
        ]

    def targets(self) -> list[torch.Tensor]:
        """Return the targets for all samples in the dataset.

        Returns
        -------
            The targets for all samples in the dataset.
        """

        return list(itertools.chain(*[k.targets() for k in self._datasets]))

    def __getitem__(self, key: int) -> Sample:
        i, j = self._indices[key]
        return self._datasets[i][j]

    def __len__(self):
        return sum([len(k) for k in self._datasets])

    def __iter__(self):
        for dataset in self._datasets:
            yield from dataset


class ConcatDataModule(lightning.LightningDataModule):
    """A conveninent DataModule with dictionary split loading, mini- batching,
    parallelisation and caching, all in one.

    Instances of this class can load and concatenate an arbitrary number of
    data-split (a.k.a. protocol) definitions for (possibly disjoint) databases,
    and can manage raw data-loading from disk.  An optional caching mechanism
    stores the data in associated CPU memory, which can improve data serving
    while training and evaluating models.

    This DataModule defines basic operations to handle data loading and
    mini-batch handling within this package's framework.  It can return
    :py:class:`torch.utils.data.DataLoader` objects for training, validation,
    prediction and testing conditions.  Parallelisation is handled by a simple
    input flag.

    Parameters
    ----------
    splits
        A dictionary that contains string keys representing dataset names, and
        values that are iterables over a 2-tuple containing an iterable over
        arbitrary, user-configurable sample representations (potentially on
        disk or permanent storage), and :py:class:`.data.typing.RawDataLoader`
        (or "sample") loader objects, which concretely implement a mechanism to
        load such samples in memory, from permanent storage.

        Sample representations on permanent storage may be of any iterable
        format (e.g. list, dictionary, etc.), for as long as the assigned
        :py:class:`.data.typing.RawDataLoader` can properly handle it.

        .. tip::

           To check the split and that the loader function works correctly, you may
           use :py:func:`.split.check_database_split_loading`.

        This class expects at least one entry called ``train`` to exist in the
        input dictionary.  Optional entries are ``validation``, and ``test``.
        Entries named ``monitor-...`` will be considered extra datasets that do
        not influence any early stop criteria during training, and are just
        monitored beyond the ``validation`` dataset.
    database_name
        The name of the database, or aggregated database containing the
        raw-samples served by this data module.
    split_name
        The name of the split used to group the samples into the various
        datasets for training, validation and testing.
    task
        The task this datamodule generate samples for (e.g. ``classification``,
        ``segmentation``, or ``detection``).
    num_classes
        The number of target classes samples of this datamodule can have. In a
        classification task, this will dictate the number of outputs for the classifier
        (one-hot-encoded), the number of segmentation outputs for a semantic
        segmentation network, or the types of objects in an object detector.
    collate_fn
        A custom function to batch the samples.
        Uses torch.utils.data.default_collate() by default.
    cache_samples
        If set, then issue raw data loading during ``prepare_data()``, and
        serves samples from CPU memory.  Otherwise, loads samples from disk on
        demand. Running from CPU memory will offer increased speeds in exchange
        for CPU memory.  Sufficient CPU memory must be available before you set
        this attribute to ``True``.  It is typically useful for relatively small
        datasets.
    batch_size
        Number of samples in every **training** batch (this parameter affects
        memory requirements for the network).  If the number of samples in the
        batch is larger than the total number of samples available for
        training, this value is truncated.  If this number is smaller, then
        batches of the specified size are created and fed to the network until
        there are no more new samples to feed (epoch is finished).  If the
        total number of training samples is not a multiple of the batch-size,
        the last batch will be smaller than the first, unless
        ``drop_incomplete_batch`` is set to ``true``, in which case this batch
        is not used.
    drop_incomplete_batch
        If set, then may drop the last batch in an epoch in case it is
        incomplete.  If you set this option, you should also consider
        increasing the total number of training epochs, as the total number
        of training steps may be reduced.
    parallel
        Use multiprocessing for data loading: if set to -1 (default), disables
        multiprocessing data loading.  Set to 0 to enable as many data loading
        instances as processing cores available in the system.  Set to >= 1
        to enable that many multiprocessing instances for data loading.
    """

    DatasetDictionary: typing.TypeAlias = dict[str, Dataset]
    """A dictionary of datasets mapping names to actual datasets."""

    def __init__(
        self,
        splits: ConcatDatabaseSplit,
        database_name: str = "",
        split_name: str = "",
        task: str = "",
        num_classes: int = 1,
        collate_fn=torch.utils.data.default_collate,
        cache_samples: bool = False,
        batch_size: int = 1,
        drop_incomplete_batch: bool = False,
        parallel: int = -1,
    ):
        super().__init__()

        self.splits = splits
        self.database_name = database_name
        self.split_name = split_name
        self.task = task
        self.num_classes = num_classes
        self.collate_fn = collate_fn

        for dataset_name, split_loaders in splits.items():
            count = sum([len(k) for k, _ in split_loaders])
            logger.info(
                f"Dataset `{dataset_name}` (`{database_name}`/`{split_name}`) "
                f"contains {count} samples",
            )

        self.cache_samples = cache_samples

        self._model_transforms: TransformSequence | None = None

        self.batch_size = batch_size

        self.drop_incomplete_batch = drop_incomplete_batch
        self.parallel = parallel  # immutable, otherwise would need to call

        self.pin_memory = (
            torch.cuda.is_available() or torch.backends.mps.is_available()  # type: ignore
        )  # should only be true if GPU available and using it

        # datasets that have been setup() for the current stage
        self._datasets: ConcatDataModule.DatasetDictionary = {}

    @property
    def parallel(self) -> int:
        """Whether to use multiprocessing for data loading.

        Use multiprocessing for data loading: if set to -1 (default),
        disables multiprocessing data loading.  Set to 0 to enable as
        many data loading instances as processing cores available in
        the system.  Set to >= 1 to enable that many multiprocessing
        instances for data loading.

        It sets the parameter ``num_workers`` (from DataLoaders) to match the
        expected pytorch representation.  It also sets the
        ``multiprocessing_context`` to use ``spawn`` instead of the default
        (``fork``, on Linux).

        The mapping between the command-line interface ``parallel`` setting
        works like this:

        .. list-table:: Relationship between ``parallel`` and DataLoader parameters
           :widths: 15 15 70
           :header-rows: 1

           * - CLI ``parallel``
             - :py:class:`torch.utils.data.DataLoader` ``kwargs``
             - Comments
           * - ``<0``
             - 0
             - Disables multiprocessing entirely, executes everything within the
               same processing context
           * - ``0``
             - :py:func:`multiprocessing.cpu_count`
             - Runs mini-batch data loading on as many external processes as CPUs
               available in the current machine
           * - ``>=1``
             - ``parallel``
             - Runs mini-batch data loading on as many external processes as set on
               ``parallel``

        Returns
        -------
        int
            The value of self._parallel.
        """
        return self._parallel

    @parallel.setter
    def parallel(self, value: int) -> None:
        self._dataloader_multiproc: dict[str, typing.Any] = {}
        self._parallel = value

        if value < 0:
            num_workers = 0

        else:
            num_workers = value or torch.multiprocessing.cpu_count()

        self._dataloader_multiproc["num_workers"] = num_workers

        if num_workers > 0:
            self._dataloader_multiproc["multiprocessing_context"] = "spawn"

        # keep workers hanging around if we have multiple
        if value >= 0:
            self._dataloader_multiproc["persistent_workers"] = True

    @property
    def model_transforms(self) -> TransformSequence | None:
        """Transform required to fit data into the model.

        A list of transforms (torch modules) that will be applied after
        raw-data-loading. and just before data is fed into the model or
        eventual data-augmentation transformations for all data loaders
        produced by this DataModule.  This part of the pipeline receives
        data as output by the raw-data-loader, or model-related
        transforms (e.g. resize adaptions), if any is specified.  If
        data is cached, it is cached **after** model-transforms are
        applied, as that is a potential memory saver (e.g., if it
        contains a resizing operation to smaller images).

        Returns
        -------
        list
            A list containing the model tansforms.
        """
        return self._model_transforms

    @model_transforms.setter
    def model_transforms(self, value: TransformSequence | None):
        old_value = self._model_transforms
        if value is None:
            self._model_transforms = value
        else:
            self._model_transforms = list(value)

        # datasets that have been setup() for the current stage are reset
        if value != old_value and len(self._datasets):
            logger.warning(
                f"Resetting {len(self._datasets)} loaded datasets due "
                "to changes in model-transform properties.  If you were caching "
                "data loading, this will (eventually) trigger a reload.",
            )
            self._datasets = {}

    def _setup_dataset(self, name: str) -> None:
        """Set up a single dataset from the input data split.

        Parameters
        ----------
        name
            Name of the dataset to setup.
        """

        if self.model_transforms is None:
            raise RuntimeError(
                "Parameter `model_transforms` has not yet been "
                "set.  If you do not have model transforms, then "
                "set it to an empty list.",
            )

        if name in self._datasets:
            logger.info(
                f"Dataset `{name}` is already setup. " f"Not re-instantiating it.",
            )
            return

        datasets: list[CachedDataset | _DelayedLoadingDataset] = []
        if self.cache_samples:
            logger.info(
                f"Loading dataset:`{name}` into memory (caching)."
                f" Trade-off: CPU RAM usage: more | Disk I/O: less",
            )
            for split, loader in self.splits[name]:
                datasets.append(
                    CachedDataset(
                        split,
                        loader,
                        self.model_transforms,
                        self.parallel,
                        self._dataloader_multiproc.get("multiprocessing_context"),
                    ),
                )
        else:
            logger.info(
                f"Loading dataset:`{name}` without caching."
                f" Trade-off: CPU RAM usage: less | Disk I/O: more",
            )
            for split, loader in self.splits[name]:
                datasets.append(
                    _DelayedLoadingDataset(
                        split,
                        loader,
                        self.model_transforms,
                    ),
                )

        if len(datasets) == 1:
            self._datasets[name] = datasets[0]
        else:
            self._datasets[name] = ConcatDataset(datasets)

    def val_dataset_keys(self) -> list[str]:
        """Return list of validation dataset names.

        Returns
        -------
        list[str]
            The list of validation dataset names.
        """

        validation_split_name = "validation"
        if "validation" not in self.splits.keys():
            logger.warning(
                "No split named 'validation', the training split will be used for validation instead."
            )
            validation_split_name = "train"

        return [validation_split_name] + [
            k for k in self.splits.keys() if k.startswith("monitor-")
        ]

    def setup(self, stage: str) -> None:
        """Set up datasets for different tasks on the pipeline.

        This method should setup (load, pre-process, etc) all datasets required
        for a particular ``stage`` (fit, validate, test, predict), and keep
        them ready to be used on one of the `_dataloader()` functions that are
        pertinent for such stage.

        If you have set ``cache_samples``, samples are loaded at this stage and
        cached in memory.

        Parameters
        ----------
        stage
            Name of the stage in which the setup is applicable.  Can be one of
            ``fit``, ``validate``, ``test`` or ``predict``.  Each stage
            typically uses the following data loaders:

            * ``fit``: uses both train and validation datasets
            * ``validate``: uses only the validation dataset
            * ``test``: uses only the test dataset
            * ``predict``: uses only the test dataset
        """

        if stage == "fit":
            for k in ["train"] + self.val_dataset_keys():
                self._setup_dataset(k)

        elif stage == "validate":
            for k in self.val_dataset_keys():
                self._setup_dataset(k)

        elif stage == "test":
            self._setup_dataset("test")

        elif stage == "predict":
            for k in self.splits:
                self._setup_dataset(k)

    def teardown(self, stage: str) -> None:
        """Unset-up datasets for different tasks on the pipeline.

        This method unsets (unload, remove from memory, etc) all datasets required
        for a particular ``stage`` (fit, validate, test, predict).

        If you have set ``cache_samples``, samples are loaded and this may
        effectivley release all the associated memory.

        Parameters
        ----------
        stage
            Name of the stage in which the teardown is applicable.  Can be one of
            ``fit``, ``validate``, ``test`` or ``predict``.  Each stage
            typically uses the following data loaders:

            * ``fit``: uses both train and validation datasets
            * ``validate``: uses only the validation dataset
            * ``test``: uses only the test dataset
            * ``predict``: uses only the test dataset
        """

        super().teardown(stage)

    def train_dataloader(self) -> DataLoader:
        """Return the train data loader.

        Returns
        -------
            The train data loader(s).
        """

        return torch.utils.data.DataLoader(
            self._datasets["train"],
            shuffle=True,
            batch_size=self.batch_size,
            drop_last=self.drop_incomplete_batch,
            pin_memory=self.pin_memory,
            collate_fn=self.collate_fn,
            **self._dataloader_multiproc,
        )

    def unshuffled_train_dataloader(self) -> DataLoader:
        """Return the train data loader without shuffling.

        Returns
        -------
            The train data loader without shuffling.
        """

        return torch.utils.data.DataLoader(
            self._datasets["train"],
            shuffle=False,
            batch_size=self.batch_size,
            drop_last=False,
            collate_fn=self.collate_fn,
            **self._dataloader_multiproc,
        )

    def val_dataloader(self) -> dict[str, DataLoader]:
        """Return the validation data loader(s).

        Returns
        -------
            The validation data loader(s).
        """

        validation_loader_opts = {
            "batch_size": self.batch_size,
            "shuffle": False,
            "drop_last": self.drop_incomplete_batch,
            "pin_memory": self.pin_memory,
        }
        validation_loader_opts.update(self._dataloader_multiproc)

        return {
            k: torch.utils.data.DataLoader(
                self._datasets[k],
                collate_fn=self.collate_fn,
                **validation_loader_opts,
            )
            for k in self.val_dataset_keys()
        }

    def test_dataloader(self) -> dict[str, DataLoader]:
        """Return the test data loader(s).

        Returns
        -------
            The test data loader(s).
        """

        return dict(
            test=torch.utils.data.DataLoader(
                self._datasets["test"],
                batch_size=self.batch_size,
                shuffle=False,
                drop_last=self.drop_incomplete_batch,
                pin_memory=self.pin_memory,
                collate_fn=self.collate_fn,
                **self._dataloader_multiproc,
            ),
        )

    def predict_dataloader(self) -> dict[str, DataLoader]:
        """Return the prediction data loader(s).

        Returns
        -------
            The prediction data loader(s).
        """

        return {
            k: torch.utils.data.DataLoader(
                self._datasets[k],
                batch_size=self.batch_size,
                shuffle=False,
                drop_last=self.drop_incomplete_batch,
                pin_memory=self.pin_memory,
                collate_fn=self.collate_fn,
                **self._dataloader_multiproc,
            )
            for k in self._datasets
        }


class CachingDataModule(ConcatDataModule):
    """A simplified version of our DataModule for a single split.

    Apart from construction, the behaviour of this DataModule is very similar
    to its simpler counterpart, serving training, validation and test sets.

    Parameters
    ----------
    database_split
        A dictionary that contains string keys representing dataset names, and
        values that are iterables over sample representations (potentially on
        disk).  These objects are passed to an unique
        :py:class:`.data.typing.RawDataLoader` for loading the
        :py:data:`.typing.Sample` data (and metadata) in memory.  It
        therefore assumes the whole split is homogeneous and can be loaded in
        the same way.

        .. tip::

           To check the split and the loader function works correctly, you may
           use :py:func:`.split.check_database_split_loading`.

        This class expects at least one entry called ``train`` to exist in the
        input dictionary.  Optional entries are ``validation``, and ``test``.
        Entries named ``monitor-...`` will be considered extra datasets that do
        not influence any early stop criteria during training, and are just
        monitored beyond the ``validation`` dataset.
    raw_data_loader
        An object instance that can load samples from storage.
    **kwargs
        List of named parameters matching those of
        :py:class:`ConcatDataModule`, other than ``splits``.
    """

    def __init__(
        self,
        database_split: DatabaseSplit,
        raw_data_loader: RawDataLoader,
        **kwargs,
    ):
        splits = {k: [(v, raw_data_loader)] for k, v in database_split.items()}
        super().__init__(
            splits=splits,
            **kwargs,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Tests for measure functions."""

import numpy


def test_centered_maxf1() -> None:
    from mednet.engine.classify.evaluator import _get_centered_maxf1

    # Multiple max F1
    f1_scores = numpy.array([0.8, 0.9, 1.0, 1.0, 1.0, 0.3])
    thresholds = numpy.array([0.2, 0.3, 0.4, 0.5, 0.6, 0.7])

    maxf1, threshold = _get_centered_maxf1(f1_scores, thresholds)

    assert maxf1 == 1.0
    assert threshold == 0.5

    # Single max F1
    f1_scores = numpy.array([0.8, 0.9, 1.0, 0.9, 0.7, 0.3])
    thresholds = numpy.array([0.2, 0.3, 0.4, 0.5, 0.6, 0.7])

    maxf1, threshold = _get_centered_maxf1(f1_scores, thresholds)

    assert maxf1 == 1.0
    assert threshold == 0.4


def test_run_binary_1() -> None:
    from mednet.engine.classify.evaluator import run
    from mednet.models.classify.typing import Prediction

    predictions: list[Prediction] = [
        # (name, target, predicted-value)
        ("s0", [0], [0.1]),
        ("s2", [0], [0.8]),
        ("s3", [1], [0.9]),
        ("s3", [1], [0.4]),
    ]

    results = run(
        "test",
        predictions,
        binning=10,
        threshold_a_priori=0.5,
    )

    assert results["num_samples"] == 4
    assert numpy.isclose(results["threshold"], 0.5)
    assert not results["threshold_a_posteriori"]
    assert numpy.isclose(results["precision"], 1 / 2)  # tp / (tp + fp)
    assert numpy.isclose(results["recall"], 1 / 2)  # tp / (tp + fn)
    assert numpy.isclose(
        results["f1"],
        2 * (1 / 2 * 1 / 2) / (1 / 2 + 1 / 2),
    )  # 2 * (prec. * recall) / (prec. + recall)
    assert numpy.isclose(
        results["accuracy"],
        (1 + 1) / (1 + 1 + 1 + 1),
    )  # (tp + tn) / (tp + fn + tn + fp)
    assert numpy.isclose(results["specificity"], 1 / 2)  # tn / (tn + fp)

    # threshold table:
    # threshold |  TNR  | 1-TNR |  TPR
    # ----------+-------+-------+---------
    #  < 0.1    |  0    |  1    |  1
    #    0.1    |  0.5  |  0.5  |  1
    #    0.4    |  0.5  |  0.5  |  0.5
    #    0.8    |  1    |  0    |  0.5
    #    0.9    |  1    |  0    |  0
    #  > 0.9    |  1    |  0    |  0
    assert numpy.isclose(results["roc_auc"], 0.75)

    # threshold table:
    # threshold |  Prec.  |  Recall
    # ----------+---------+----------
    #  < 0.1    |  0.5    |  1
    #    0.1    |  2/3    |  1
    #    0.4    |  0.5    |  0.5
    #    0.8    |  1      |  0.5
    #    0.9    |  0      |  0
    #  > 0.9    |  0      |  0
    assert numpy.isclose(results["average_precision"], 0.8333333)


def test_run_binary_2() -> None:
    from mednet.engine.classify.evaluator import run
    from mednet.models.classify.typing import Prediction

    predictions: list[Prediction] = [
        # (name, target, predicted-value)
        ("s0", [0], [0.1]),
        ("s2", [0], [0.8]),
        ("s3", [1], [0.9]),
        ("s3", [1], [0.4]),
    ]

    # a change in the threshold should not affect auc and average precision scores
    results = run(
        "test",
        predictions,
        binning=10,
        threshold_a_priori=0.3,
    )

    assert results["num_samples"] == 4
    assert numpy.isclose(results["threshold"], 0.3)
    assert not results["threshold_a_posteriori"]

    assert numpy.isclose(results["precision"], 2 / 3)  # tp / (tp + fp)
    assert numpy.isclose(results["recall"], 2 / 2)  # tp / (tp + fn)
    assert numpy.isclose(
        results["f1"],
        2 * (2 / 3 * 2 / 2) / (2 / 3 + 2 / 2),
    )  # 2 * (prec. * recall) / (prec. + recall)
    assert numpy.isclose(
        results["accuracy"],
        (2 + 1) / (2 + 0 + 1 + 1),
    )  # (tp + tn) / (tp + fn + tn + fp)
    assert numpy.isclose(results["specificity"], 1 / (1 + 1))  # tn / (tn + fp)

    # threshold table:
    # threshold |  TNR  | 1-TNR |  TPR
    # ----------+-------+-------+---------
    #  < 0.1    |  0    |  1    |  1
    #    0.1    |  0.5  |  0.5  |  1
    #    0.4    |  0.5  |  0.5  |  0.5
    #    0.8    |  1    |  0    |  0.5
    #    0.9    |  1    |  0    |  0
    #  > 0.9    |  1    |  0    |  0
    assert numpy.isclose(results["roc_auc"], 0.75)

    # threshold table:
    # threshold |  Prec.  |  Recall
    # ----------+---------+----------
    #  < 0.1    |  0.5    |  1
    #    0.1    |  2/3    |  1
    #    0.4    |  0.5    |  0.5
    #    0.8    |  1      |  0.5
    #    0.9    |  0      |  0
    #  > 0.9    |  0      |  0
    assert numpy.isclose(results["average_precision"], 0.8333333)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Image transformations for our pipelines.

Differences between methods here and those from
:py:mod:`torchvision.transforms` is that these support multiple
simultaneous image inputs, which are required to feed segmentation
networks (e.g. image and labels or masks).  We also take care of data
augmentations, in which random flipping and rotation needs to be applied
across all input images, but color jittering, for example, only on the
input image.
"""

import functools
import logging
import typing

import numpy.random
import numpy.typing
import torch
from scipy.ndimage import gaussian_filter, map_coordinates

logger = logging.getLogger(__name__)


def _elastic_deformation_on_image(
    img: torch.Tensor,
    rng: numpy.random.Generator,
    alpha: float = 1000.0,
    sigma: float = 30.0,
    spline_order: int = 1,
    mode: str = "nearest",
    p: float = 1.0,
) -> torch.Tensor:
    """Perform elastic deformation on an image.

    This implementation is based on 2 scipy functions
    (:py:func:`scipy.ndimage.gaussian_filter` and
    :py:func:`scipy.ndimage.map_coordinates`).  It is very inefficient since it
    requires data to be moved off the current running device and then back.

    Parameters
    ----------
    img
        The input image to apply elastic deformation to.  This image should
        always have this shape: ``[C, H, W]``. It should always represent a
        tensor on the CPU.
    rng
        The numpy random number generator to use.
    alpha
        A multiplier for the gaussian filter outputs.
    sigma
        Standard deviation for Gaussian kernel.
    spline_order
        The order of the spline interpolation, default is 1. The order has to
        be in the range 0-5.
    mode
        The mode parameter determines how the input array is extended beyond
        its boundaries.
    p
        Probability that this transformation will be applied.  Meaningful when
        using it as a data augmentation technique.

    Returns
    -------
    tensor
        The image with elastic deformation applied, as a tensor on the CPU.
    """

    if rng.random() < p:
        assert img.ndim == 3, (
            f"This filter accepts only images with 3 dimensions, "
            f"however I got an image with {img.ndim} dimensions."
        )

        # Input tensor is of shape C x H x W
        img_shape = img.shape[1:]

        dx = alpha * typing.cast(
            numpy.typing.NDArray[numpy.float64],
            gaussian_filter(
                (rng.random(size=(img_shape[0], img_shape[1])) * 2 - 1),
                sigma,
                mode="constant",
                cval=0.0,
            ),
        )
        dy = alpha * typing.cast(
            numpy.typing.NDArray[numpy.float64],
            gaussian_filter(
                (rng.random(size=(img_shape[0], img_shape[1])) * 2 - 1),
                sigma,
                mode="constant",
                cval=0.0,
            ),
        )

        x, y = numpy.meshgrid(
            numpy.arange(img_shape[0]),
            numpy.arange(img_shape[1]),
            indexing="ij",
        )
        indices = [
            numpy.reshape(x + dx, (-1, 1)),
            numpy.reshape(y + dy, (-1, 1)),
        ]

        # may copy, if img is not on CPU originally
        img_numpy = img.numpy()
        output = numpy.zeros_like(img_numpy)
        for i in range(img.shape[0]):
            output[i, :, :] = torch.tensor(
                map_coordinates(
                    img_numpy[i, :, :],
                    indices,
                    order=spline_order,
                    mode=mode,
                ).reshape(img_shape),
            )

        # wraps numpy array as tensor, move to destination device if need-be
        return torch.as_tensor(output)

    return img


def _elastic_deformation_on_batch(
    batch: torch.Tensor,
    rng: numpy.random.Generator,
    alpha: float = 1000.0,
    sigma: float = 30.0,
    spline_order: int = 1,
    mode: str = "nearest",
    p: float = 1.0,
) -> torch.Tensor:
    """Perform elastic deformation on a batch of images.

    This implementation is based on 2 scipy functions
    (:py:func:`scipy.ndimage.gaussian_filter` and
    :py:func:`scipy.ndimage.map_coordinates`).  It is very inefficient since it
    requires data to be moved off the current running device and then back.

    Parameters
    ----------
    batch
        The batch to apply elastic deformation to.
    rng
        The numpy random number generator to use.
    alpha
        A multiplier for the gaussian filter outputs.
    sigma
        Standard deviation for Gaussian kernel.
    spline_order
        The order of the spline interpolation, default is 1. The order has to
        be in the range 0-5.
    mode
        The mode parameter determines how the input array is extended beyond
        its boundaries.
    p
        Probability that this transformation will be applied.  Meaningful when
        using it as a data augmentation technique.

    Returns
    -------
    tensor
        A batch of images with elastic deformation applied, as a tensor on the CPU.
    """

    # transforms our custom functions into simpler callables
    partial = functools.partial(
        _elastic_deformation_on_image,
        rng=rng,
        alpha=alpha,
        sigma=sigma,
        spline_order=spline_order,
        mode=mode,
        p=p,
    )

    # if a mp pool is available, do it in parallel
    augmented_images: typing.Any
    augmented_images = map(partial, batch.cpu())
    return torch.stack(list(augmented_images))


class ElasticDeformation:
    """Elastic deformation of 2D image slightly adapted from :cite:p:`simard_best_2003`.

    This implementation is based on 2 scipy functions
    (:py:func:`scipy.ndimage.gaussian_filter` and
    :py:func:`scipy.ndimage.map_coordinates`).  It is very inefficient since it
    requires data to be moved off the current running device and then back.

    .. warning::

       Furthermore, this transform is not scriptable and therefore cannot run
       on a CUDA or MPS device.  Applying it effectively creates a bottleneck
       in model training.

    Source: https://gist.github.com/oeway/2e3b989e0343f0884388ed7ed82eb3b0

    Parameters
    ----------
    alpha
        A multiplier for the gaussian filter outputs.
    sigma
        Standard deviation for Gaussian kernel.
    spline_order
        The order of the spline interpolation, default is 1. The order has to
        be in the range 0-5.
    mode
        The mode parameter determines how the input array is extended beyond
        its boundaries.
    p
        Probability that this transformation will be applied.  Meaningful when
        using it as a data augmentation technique.
    seed
        Set the random generator seed, if given.  Otherwise, initializes the generator
        with a random seed (c.f. :py:func:`numpy.random.default_rng`).
    """

    def __init__(
        self,
        alpha: float = 1000.0,
        sigma: float = 30.0,
        spline_order: int = 1,
        mode: str = "nearest",
        p: float = 1.0,
        seed: int | None = None,
    ):
        self.alpha: float = alpha
        self.sigma: float = sigma
        self.spline_order: int = spline_order
        self.mode: str = mode
        self.p: float = p
        self.seed = seed
        self.npy_rng = numpy.random.default_rng(seed=seed)

    def __str__(self) -> str:
        parameters = [
            f"alpha={self.alpha}",
            f"sigma={self.sigma}",
            f"spline_order={self.spline_order}",
            f"mode={self.mode}",
            f"p={self.p}",
            f"seed={self.seed}",
        ]
        return f"{type(self).__name__}({', '.join(parameters)})"

    def __call__(self, img: torch.Tensor) -> torch.Tensor:
        if len(img.shape) == 4:
            return _elastic_deformation_on_batch(
                batch=img,
                rng=self.npy_rng,
                alpha=self.alpha,
                sigma=self.sigma,
                spline_order=self.spline_order,
                mode=self.mode,
                p=self.p,
            ).to(img.device)

        if len(img.shape) == 3:
            return _elastic_deformation_on_image(
                img=img.cpu(),
                rng=self.npy_rng,
                alpha=self.alpha,
                sigma=self.sigma,
                spline_order=self.spline_order,
                mode=self.mode,
                p=self.p,
            ).to(img.device)

        raise RuntimeError(
            f"This transform accepts only images with 3 dimensions,"
            f"or batches of images with 4 dimensions.  However, I got "
            f"an image with {img.ndim} dimensions.",
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""TB-POC dataset for computer-aided diagnosis.

This databases contain only the tuberculosis final diagnosis (0 or 1) and come
from HIV infected patients.

* Database reference: :cite:p:`griesel_optimizing_2018`

.. important:: **Raw data organization**

    The TB-POC base datadir, which you should configure following the
    :ref:`mednet.setup` instructions, must contain at least the directory
    ``TBPOC_CXR`` with all JPEG images.

Data specifications:

* Raw data input (on disk):

  * JPEG 8-bit Grayscale images
  * Original resolution (height x width or width x height): 2048 x 2500 pixels
    or 2500 x 2048 pixels
  * Total samples: 407

* Output image:

  * Transforms:

    * Load raw grayscale jpeg with :py:mod:`PIL`
    * Remove black borders
    * Convert to torch tensor
    * Torch center cropping to get square image

  * Final specifications:

    * Grayscale, encoded as a single plane tensor, 32-bit floats,
      square with varying resolutions (2048 x 2048 being the maximum),
      but also depending on black borders' sizes on the input image.
    * Labels: 0 (healthy), 1 (active tuberculosis)

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources.abc
import os
import pathlib
import typing

import PIL.Image
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..image_utils import remove_black_borders
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name of this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the Shenzen dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = pathlib.Path(
            load_rc().get(
                CONFIGURATION_KEY_DATADIR,
                os.path.realpath(os.curdir),
            ),
        )

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The sample representation.
        """

        # images from TBPOC are encoded as grayscale JPEGs, no need to
        # call convert("L") here.
        image = PIL.Image.open(self.datadir / sample[0])
        image, _ = remove_black_borders(image)
        image = to_dtype(to_image(image), torch.float32, scale=True)
        image = tv_tensors.Image(image)

        # use the code below to view generated images
        # from torchvision.transforms.v2.functional import to_pil_image
        # to_pil_image(tensor).show()
        # __import__("pdb").set_trace()

        return dict(image=image, target=self.target(sample), name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            1D torch float tensor.
        """

        return torch.FloatTensor([sample[1]])


class DataModule(CachingDataModule):
    """TB-POC dataset for computer-aided diagnosis.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="classification",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import pathlib
import typing

import h5py
import lightning.pytorch
import lightning.pytorch.callbacks
import torch.utils.data
import tqdm

from ...engine.device import DeviceManager
from ...utils.string import rewrap

logger = logging.getLogger(__name__)


class _HDF5Writer(lightning.pytorch.callbacks.BasePredictionWriter):
    """Write HDF5 files for each sample processed by our model.

    Objects of this class can also keep track of samples written to disk and
    return a summary list.

    Parameters
    ----------
    output_folder
        Base directory where to write predictions to.

    write_interval
        When will this callback be active.
    """

    def __init__(
        self,
        output_folder: pathlib.Path,
        write_interval: typing.Literal["batch", "epoch", "batch_and_epoch"] = "batch",
    ):
        super().__init__(write_interval=write_interval)
        self.output_folder = output_folder
        self._data: list[tuple[str, str]] = []

    def write_on_batch_end(
        self,
        trainer: lightning.pytorch.Trainer,
        pl_module: lightning.pytorch.LightningModule,
        prediction: typing.Any,
        batch_indices: typing.Sequence[int] | None,
        batch: typing.Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        """Write batch predictions to disk.

        Parameters
        ----------
        trainer
            The trainer being used.
        pl_module
            The pytorch module.
        prediction
            The actual predictions to record.
        batch_indices
            The relative position of samples on the epoch.
        batch
            The current batch.
        batch_idx
            Index of the batch overall.
        dataloader_idx
            Index of the dataloader overall.
        """
        del trainer, pl_module, batch_indices, batch_idx, dataloader_idx

        for k, sample_pred in enumerate(prediction):
            sample_name: str = batch["name"][k]
            stem = pathlib.Path(sample_name).with_suffix(".hdf5")
            output_path = self.output_folder / stem
            tqdm.tqdm.write(f"`{sample_name}` -> `{str(output_path)}`")
            output_path.parent.mkdir(parents=True, exist_ok=True)
            with h5py.File(output_path, "w") as f:
                f.create_dataset(
                    "image",
                    data=batch["image"][k].cpu().numpy(),
                    compression="gzip",
                    compression_opts=9,
                )
                f.create_dataset(
                    "prediction",
                    data=sample_pred.cpu().numpy().squeeze(0),
                    compression="gzip",
                    compression_opts=9,
                )
                f.create_dataset(
                    "target",
                    data=(batch["target"][k].squeeze(0).cpu().numpy() > 0.5),
                    compression="gzip",
                    compression_opts=9,
                )
                f.create_dataset(
                    "mask",
                    data=(batch["mask"][k].squeeze(0).cpu().numpy() > 0.5),
                    compression="gzip",
                    compression_opts=9,
                )
            self._data.append((sample_name, str(stem)))

    def reset(self) -> list[tuple[str, str]]:
        """Summary of written objects.

        Also resets the internal state.

        Returns
        -------
            A list containing a summary of all samples written.
        """
        retval = self._data
        self._data = []
        return retval


def run(
    model: lightning.pytorch.LightningModule,
    datamodule: lightning.pytorch.LightningDataModule,
    device_manager: DeviceManager,
    output_folder: pathlib.Path,
) -> (
    dict[str, list[tuple[str, str]]]
    | list[list[tuple[str, str]]]
    | list[tuple[str, str]]
    | None
):
    """Run inference on input data, output predictions.

    Parameters
    ----------
    model
        Neural network model (e.g. lwnet).
    datamodule
        The lightning DataModule to run predictions on.
    device_manager
        An internal device representation, to be used for prediction. This
        representation can be converted into a pytorch device or a lightning
        accelerator setup.
    output_folder
        Folder where to store HDF5 representations of probability maps.

    Returns
    -------
        A JSON-able representation of sample data stored at ``output_folder``.
        For every split (dataloader), a list of samples in the form
        ``[sample-name, hdf5-path]`` is returned.  In the cases where the
        ``predict_dataloader()`` returns a single loader, we then return a
        list.  A dictionary is returned in case ``predict_dataloader()`` also
        returns a dictionary.

    Raises
    ------
    TypeError
        If the DataModule's ``predict_dataloader()`` method does not return any
        of the types described above.
    """

    from lightning.pytorch.loggers.logger import DummyLogger

    collector = _HDF5Writer(output_folder)

    accelerator, devices = device_manager.lightning_accelerator()
    trainer = lightning.pytorch.Trainer(
        accelerator=accelerator,
        devices=devices,
        logger=DummyLogger(),
        callbacks=[collector],
    )

    dataloaders = datamodule.predict_dataloader()

    if isinstance(dataloaders, torch.utils.data.DataLoader):
        logger.info("Running prediction on a single dataloader...")
        trainer.predict(model, dataloaders, return_predictions=False)
        return collector.reset()

    if isinstance(dataloaders, list):
        retval_list = []
        for k, dataloader in enumerate(dataloaders):
            logger.info(f"Running prediction on split `{k}`...")
            trainer.predict(model, dataloader, return_predictions=False)
            retval_list.append(collector.reset())
        return retval_list

    if isinstance(dataloaders, dict):
        retval_dict = {}
        for name, dataloader in dataloaders.items():
            logger.info(f"Running prediction on `{name}` split...")
            trainer.predict(model, dataloader, return_predictions=False)
            retval_dict[name] = collector.reset()
        return retval_dict

    if dataloaders is None:
        logger.warning("Datamodule did not return any prediction dataloaders!")
        return None

    # if you get to this point, then the user is returning something that is
    # not supported - complain!
    raise TypeError(
        rewrap(
            f"""Datamodule returned strangely typed prediction dataloaders:
            `{type(dataloaders)}` - if this is not an error, write code to support this
            use-case."""
        )
    )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""VISCERAL dataset for 3D organ classification (only lungs and bladders).

* Database reference: :cite:p:`jimenez-del-toro_cloud-based_2016`

Data specifications:

* Raw data input (on disk):

  * NIfTI volumes
  * resolution: 16x16x16 pixels - Loaded samples are not full scans but
    16x16x16 volumes of organs.

* Output image:

  * Transforms:

    * Load raw NIfTI with `torchio <https://torchio.readthedocs.io/>`_
    * Clamp and Rescale intensity
    * Convert to torch tensor

  * Final specifications

    * 32-bit floats, cubes 16x16x16 pixels
    * targets: 0 (bladder), 1 (lung)

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources.abc
import os
import pathlib
import typing

import torch
import torchio as tio
from torchvision import tv_tensors

from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name of this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the VISCERAL database."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self) -> None:
        self.datadir = pathlib.Path(
            load_rc().get(
                CONFIGURATION_KEY_DATADIR,
                os.path.realpath(os.curdir),
            ),
        )

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single volume sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the database root folder,
            where to find the volume to be loaded and an integer, representing
            the sample target.

        Returns
        -------
            The sample representation.
        """
        clamp = tio.Clamp(out_min=-1000, out_max=2000)
        rescale = tio.RescaleIntensity(percentiles=(0.5, 99.5))
        preprocess = tio.Compose([clamp, rescale])
        image = tio.ScalarImage(self.datadir / sample[0])
        image = preprocess(image)
        image = tv_tensors.Image(image.data)
        return dict(image=image, target=self.target(sample), name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            1D torch float tensor.
        """

        return torch.FloatTensor([sample[1]])


class DataModule(CachingDataModule):
    """VISCERAL DataModule for 3D organ binary classification.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="classification",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
""":py:mod:`TBX11k database <mednet.data.classify.tbx11k>` (v2: healthy, sick and latent vs active TB; cross validation fold 3)."""

import importlib.resources

from mednet.data.classify.tbx11k import DataModule

datamodule = DataModule(
    importlib.resources.files(__package__ or __name__.rsplit(".", 1)[0])
    / "v2-fold-3.json"
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import torch
import torch.nn
from torch.nn import Conv2d, ConvTranspose2d


def conv_with_kaiming_uniform(
    in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1
):
    """Convolution layer with kaiming uniform.

    Parameters
    ----------
    in_channels
        Number of input channels.
    out_channels
        Number of output channels.
    kernel_size
        The kernel size.
    stride
        The stride.
    padding
        The padding.
    dilation
        The dilation.

    Returns
    -------
        The convoluation layer.
    """
    conv = Conv2d(
        in_channels,
        out_channels,
        kernel_size=kernel_size,
        stride=stride,
        padding=padding,
        dilation=dilation,
        bias=True,
    )
    # Caffe2 implementation uses XavierFill, which in fact
    # corresponds to kaiming_uniform_ in PyTorch
    torch.nn.init.kaiming_uniform_(conv.weight, a=1)
    torch.nn.init.constant_(conv.bias, 0)
    return conv


def convtrans_with_kaiming_uniform(
    in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1
):
    """Implement convtrans layer with kaiming uniform.

    Parameters
    ----------
    in_channels
        Number of input channels.
    out_channels
        Number of output channels.
    kernel_size
        The kernel size.
    stride
        The stride.
    padding
        The padding.
    dilation
        The dilation.

    Returns
    -------
        The convtrans layer.
    """
    conv = ConvTranspose2d(
        in_channels,
        out_channels,
        kernel_size=kernel_size,
        stride=stride,
        padding=padding,
        dilation=dilation,
        bias=True,
    )
    # Caffe2 implementation uses XavierFill, which in fact
    # corresponds to kaiming_uniform_ in PyTorch
    torch.nn.init.kaiming_uniform_(conv.weight, a=1)
    torch.nn.init.constant_(conv.bias, 0)
    return conv


class UpsampleCropBlock(torch.nn.Module):
    """Combines Conv2d, ConvTransposed2d and Cropping. Simulates the caffe2
    crop layer in the forward function.

    Used for DRIU and HED.

    Parameters
    ----------
    in_channels
        Number of channels of intermediate layer.
    out_channels
        Number of output channels.
    up_kernel_size
        Kernel size for transposed convolution.
    up_stride
        Stride for transposed convolution.
    up_padding
        Padding for transposed convolution.
    pixelshuffle
        If True, uses PixelShuffleICNR upsampling.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        up_kernel_size: int,
        up_stride: int,
        up_padding: int,
        pixelshuffle: bool = False,
    ):
        super().__init__()
        # NOTE: Kaiming init, replace with torch.nn.Conv2d and
        # torch.nn.ConvTranspose2d to get original DRIU impl.
        self.conv = conv_with_kaiming_uniform(in_channels, out_channels, 3, 1, 1)
        if pixelshuffle:
            self.upconv = PixelShuffleICNR(out_channels, out_channels, scale=up_stride)
        else:
            self.upconv = convtrans_with_kaiming_uniform(
                out_channels,
                out_channels,
                up_kernel_size,
                up_stride,
                up_padding,
            )

    def forward(self, x, input_res):
        img_h = input_res[0]
        img_w = input_res[1]
        x = self.conv(x)
        x = self.upconv(x)
        # determine center crop
        # height
        up_h = x.shape[2]
        h_crop = up_h - img_h
        h_s = h_crop // 2
        h_e = up_h - (h_crop - h_s)
        # width
        up_w = x.shape[3]
        w_crop = up_w - img_w
        w_s = w_crop // 2
        w_e = up_w - (w_crop - w_s)
        # perform crop
        # needs explicit ranges for onnx export
        return x[:, :, h_s:h_e, w_s:w_e]  # crop to input size


def ifnone(a, b):
    """Return ``a`` if ``a`` is not None, otherwise ``b``.

    Parameters
    ----------
    a
        The first parameter.
    b
        The second parameter.

    Returns
    -------
        The parameter a if it is not None, else b.
    """
    return b if a is None else a


def icnr(x, scale=2, init=torch.nn.init.kaiming_normal_):
    """ICNR init of ``x``, with ``scale`` and ``init`` function.

    https://docs.fast.ai/layers.html#PixelShuffleICNR.

    Parameters
    ----------
    x
        Tensor.
    scale
        Scale of the upsample.
    init
        Function used to initialize.
    """

    ni, nf, h, w = x.shape
    ni2 = int(ni / (scale**2))
    k = init(torch.zeros([ni2, nf, h, w])).transpose(0, 1)
    k = k.contiguous().view(ni2, nf, -1)
    k = k.repeat(1, 1, scale**2)
    k = k.contiguous().view([nf, ni, h, w]).transpose(0, 1)
    x.data.copy_(k)


class PixelShuffleICNR(torch.nn.Module):
    """Upsample by ``scale`` from ``ni`` filters to ``nf`` (default
    ``ni``), using ``torch.nn.PixelShuffle``, ``icnr`` init, and
    ``weight_norm``.

    https://docs.fast.ai/layers.html#PixelShuffleICNR.

    Parameters
    ----------
    ni
        Number of initial filters.
    nf
        Number of final filters.
    scale
        Scale of the upsample.
    """

    def __init__(self, ni: int, nf: int | None = None, scale: int = 2):
        super().__init__()
        nf = ifnone(nf, ni)
        self.conv = conv_with_kaiming_uniform(ni, nf * (scale**2), 1)
        icnr(self.conv.weight)
        self.shuf = torch.nn.PixelShuffle(scale)
        # Blurring over (h*w) kernel, from
        # "Super-Resolution using Convolutional Neural Networks without Any
        # Checkerboard Artifacts", https://arxiv.org/abs/1806.02658
        self.pad = torch.nn.ReplicationPad2d((1, 0, 1, 0))
        self.blur = torch.nn.AvgPool2d(2, stride=1)
        self.relu = torch.nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.shuf(self.relu(self.conv(x)))
        return self.blur(self.pad(x))


class UnetBlock(torch.nn.Module):
    """Unet block implementation.

    Parameters
    ----------
    up_in_c
        Number of input channels.
    x_in_c
        Number of cat channels.
    pixel_shuffle
        If True, uses a PixelShuffleICNR layer for upsampling.
    middle_block
        If True, uses a middle block for VGG based U-Net.
    """

    def __init__(self, up_in_c, x_in_c, pixel_shuffle=False, middle_block=False):
        super().__init__()
        # middle block for VGG based U-Net
        if middle_block:
            up_out_c = up_in_c
        else:
            up_out_c = up_in_c // 2
        cat_channels = x_in_c + up_out_c
        inner_channels = cat_channels // 2

        if pixel_shuffle:
            self.upsample = PixelShuffleICNR(up_in_c, up_out_c)
        else:
            self.upsample = convtrans_with_kaiming_uniform(up_in_c, up_out_c, 2, 2)
        self.convtrans1 = convtrans_with_kaiming_uniform(
            cat_channels, inner_channels, 3, 1, 1
        )
        self.convtrans2 = convtrans_with_kaiming_uniform(
            inner_channels, inner_channels, 3, 1, 1
        )
        self.relu = torch.nn.ReLU(inplace=True)

    def forward(self, up_in, x_in):
        up_out = self.upsample(up_in)
        cat_x = torch.cat([up_out, x_in], dim=1)
        x = self.relu(self.convtrans1(cat_x))
        return self.relu(self.convtrans2(x))



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""DRIU Network for Vessel Segmentation.

Deep Retinal Image Understanding (DRIU), a unified framework of retinal image
analysis that provides both retinal vessel and optic disc segmentation using
deep Convolutional Neural Networks (CNNs).

Reference: :cite:p:`maninis_deep_2016`
"""

import torch.optim
import torchvision.transforms
import torchvision.transforms.v2

import mednet.models.losses
import mednet.models.segment.driu_pix
import mednet.models.transforms

model = mednet.models.segment.driu_pix.DRIUPix(
    loss_type=mednet.models.losses.SoftJaccardAndBCEWithLogitsLoss,
    loss_arguments=dict(alpha=0.7),  # 0.7 BCE + 0.3 Jaccard
    optimizer_type=torch.optim.Adam,
    optimizer_arguments=dict(lr=0.01),
    model_transforms=[
        mednet.models.transforms.SquareCenterPad(),
        torchvision.transforms.v2.Resize(512, antialias=True),
        torchvision.transforms.v2.RGB(),
    ],
    pretrained=False,
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""`Logistic regression model <logistic-regression_>`_ for multi-class classification."""

import logging
import typing

import torch
import torch.nn

from ..typing import Checkpoint
from .model import Model

logger = logging.getLogger(__name__)


class LogisticRegression(Model):
    """`Logistic regression model <logistic-regression_>`_ for multi-class classification.

    Parameters
    ----------
    loss_type
        The loss to be used for training and evaluation.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.
    loss_arguments
        Arguments to the loss.
    optimizer_type
        The type of optimizer to use for training.
    optimizer_arguments
        Arguments to the optimizer after ``params``.
    num_classes
        Number of outputs (classes) for this model.
    input_size
        The number of inputs this classifer shall process.
    """

    def __init__(
        self,
        loss_type: type[torch.nn.Module] = torch.nn.BCEWithLogitsLoss,
        loss_arguments: dict[str, typing.Any] | None = None,
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.Adam,
        optimizer_arguments: dict[str, typing.Any] | None = {"lr": 1e-2},
        num_classes: int = 1,
        input_size: int = 14,
    ):
        super().__init__(
            name="logistic-regression",
            loss_type=loss_type,
            loss_arguments=loss_arguments,
            optimizer_type=optimizer_type,
            optimizer_arguments=optimizer_arguments,
            scheduler_type=None,
            scheduler_arguments=None,
            model_transforms=None,
            augmentation_transforms=None,
            num_classes=num_classes,
        )

        self.input_size = input_size
        self.num_classes = num_classes

    @Model.num_classes.setter  # type: ignore[attr-defined]
    def num_classes(self, v: int) -> None:
        self.linear = torch.nn.Linear(self.input_size, v)
        self._num_classes = v

    def on_load_checkpoint(self, checkpoint: Checkpoint) -> None:
        num_classes = checkpoint["state_dict"]["linear.bias"].shape[0]

        if num_classes != self.num_classes:
            logger.debug(
                f"Resetting number-of-output-classes at `{self.name}` model from "
                f"{self.num_classes} to {num_classes} while loading checkpoint."
            )
        self.num_classes = num_classes

        super().on_load_checkpoint(checkpoint)

    def forward(self, x):
        return self.linear(self.normalizer(x))



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
""":py:mod:`pad-chest database <mednet.data.classify.padchest>` (no TB/idiap split)."""

import importlib.resources

from mednet.data.classify.padchest import DataModule

datamodule = DataModule(
    importlib.resources.files(__package__ or __name__.rsplit(".", 1)[0])
    / "no-tb-idiap.json.bz2"
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Custom losses for different tasks."""

import logging
import typing

import numpy.typing
import torch
import torch.utils.data

from ..data.datamodule import ConcatDataModule
from ..data.typing import Dataset
from .typing import TaskType

logger = logging.getLogger(__name__)


def _task_type(
    targets: torch.Tensor
    | numpy.typing.NDArray
    | typing.Iterable[typing.Iterable[int]]
    | typing.Iterable[typing.Iterable[typing.Iterable[typing.Iterable[int]]]],
) -> TaskType:
    """Determine the type of task from combined targets available.

    This function will look into the provided targets of a dataset and will
    determine the type of task.

    Parameters
    ----------
    targets
        The complete target set, for the whole dataset being analyzed. This
        matrix should be ``[n, C]`` where ``n`` is the number of samples, and
        ``C`` the number of classes.  All values should be either 0 or 1.

    Returns
    -------
        The type of tas
    """

    int_targets = torch.Tensor(targets).int()

    task_type: TaskType = "classification"
    if len(int_targets.shape) > 2:
        task_type = "segmentation"

    return task_type


def _get_positive_weights_from_targets(targets: torch.Tensor) -> torch.Tensor:
    """Compute the weights of each class of a set of targets.

    This function inputs a set of targets and computes the ratio between number
    of negative and positive samples (scalar).  The weight can be used to
    adjust minimisation criteria to in cases there is a huge data imbalance.

    It returns a vector with weights (inverse counts) for each target.

    Parameters
    ----------
    targets
        A :py:class:`torch.Tensor` containing the targets, in the format
        ``[n, C]`` where ``n`` is the number of samples and ``C`` the
        number of classes.

    Returns
    -------
        The positive weight of each class in the dataset given as input.
    """

    task_type = _task_type(targets)

    if task_type == "segmentation":
        # rearranges ``targets`` vector so the problem looks like a simpler
        # classification problem where each pixel is a "separate sample"
        targets = targets.transpose(0, 2).transpose(1, 3).reshape(-1, targets.shape[1])

    positive_count = targets.sum(dim=0)
    negative_count = targets.shape[0] - positive_count
    return negative_count / positive_count


def _get_positive_weights_from_dataloader(
    dataloader: torch.utils.data.DataLoader,
) -> torch.Tensor:
    """Compute the weights of each class of a DataLoader.

    This function inputs a :py:class:`torch.utils.data.DataLoader` and computes
    the ratio between number of negative and positive samples (scalar).  The
    weight can be used to adjust minimisation criteria to in cases there is a
    huge data imbalance.

    It returns a vector with weights (inverse counts) for each target.

    Parameters
    ----------
    dataloader
        A DataLoader from which to compute the positive weights.  Entries must
        be a dictionary which must contain a ``target`` key.

    Returns
    -------
        The positive weight of each class in the dataset given as input.

    Raises
    ------
    NotImplementedError
        In the case of "multilabel" datasets, which are currently not
        supported.
    """

    if isinstance(dataloader.dataset, Dataset):
        # there is a faster way to access the targets!
        targets = dataloader.dataset.targets()
    else:
        targets = [batch["target"] for batch in dataloader]

    return _get_positive_weights_from_targets(torch.vstack(targets))


def pos_weight_for_bcewithlogitsloss(
    datamodule: ConcatDataModule,
) -> tuple[dict[str, torch.Tensor], dict[str, torch.Tensor]]:
    """Generate the ``pos_weight`` argument for losses of type :py:class:`torch.nn.BCEWithLogitsLoss`.

    This function can generate the ``pos_weight`` parameters for both train and
    validation losses given a datamodule.

    Parameters
    ----------
    datamodule
        The datamodule to probe for training and validation datasets.

    Returns
    -------
        A tuple containing the training and validation ``pos_weight``
        arguments, wrapped in a dictionary.
    """

    train_weights = _get_positive_weights_from_dataloader(
        datamodule.unshuffled_train_dataloader()
    )
    logger.info(f"train: BCEWithLogitsLoss(pos_weight={train_weights})")

    if "validation" in datamodule.val_dataloader().keys():
        validation_weights = _get_positive_weights_from_dataloader(
            datamodule.val_dataloader()["validation"]
        )
    else:
        logger.warning(
            "Datamodule does not contain a validation dataloader. "
            "The training dataloader will be used instead."
        )
        validation_weights = train_weights
    logger.info(f"validation: BCEWithLogitsLoss(weight={validation_weights})")

    return (dict(pos_weight=train_weights), dict(pos_weight=validation_weights))


class BCEWithLogitsLossWeightedPerBatch(torch.nn.Module):
    """Calculates the binary cross entropy loss for every batch.

    This loss is similar to :py:class:`torch.nn.BCEWithLogitsLoss`, except it
    updates the ``pos_weight`` (ratio between negative and positive target
    pixels) parameter for the loss term for every batch, based on the
    accumulated taget pixels for all samples in the batch.

    Implements Equation 1 in :cite:p:`maninis_deep_2016`.  The weight depends on the
    current proportion between negatives and positives in the ground-
    truth sample being analyzed.
    """

    def __init__(self):
        super().__init__()

    def forward(self, input_: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        input_
            Logits produced by the model to be evaluated, with the shape ``[n,
            c]`` (classification), or ``[n, c, h, w]`` (segmentation).
        target
            Ground-truth information with the shape  ``[n, c]``
            (classification), or ``[n, c, h, w]`` (segmentation), containing
            zeroes and ones.

        Returns
        -------
            The average loss for all input data.
        """

        # calculates the proportion of negatives to the total number of pixels
        # available in the masked region
        num_pos = target.sum()
        return torch.nn.functional.binary_cross_entropy_with_logits(
            input_,
            target,
            reduction="mean",
            pos_weight=(input_.numel() - num_pos) / num_pos,
        )


class SoftJaccardAndBCEWithLogitsLoss(torch.nn.Module):
    r"""Implement the generalized loss function of Equation (3) at :cite:p:`iglovikov_ternausnetv2_2018`.

    At the paper, authors suggest a value of :math:`\alpha = 0.7`, which we set
    as default for instances of this type.

    .. math::

       L = \alpha H + (1-\alpha)(1-J)

    J is the Jaccard distance, and H, the Binary Cross-Entropy Loss.  Our
    implementation is based on :py:class:`torch.nn.BCEWithLogitsLoss`.

    Parameters
    ----------
    alpha
        Determines the weighting of J and H. Default: ``0.7``.
    """

    def __init__(self, alpha: float = 0.7):
        super().__init__()
        self.alpha = alpha

    def forward(self, input_: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        input_
            Logits produced by the model to be evaluated, with the shape ``[n,
            c]`` (classification), or ``[n, c, h, w]`` (segmentation).
        target
            Ground-truth information with the shape  ``[n, c]``
            (classification), or ``[n, c, h, w]`` (segmentation), containing
            zeroes and ones.

        Returns
        -------
            Loss, in a single entry.
        """

        eps = 1e-8
        probabilities = torch.sigmoid(input_)
        intersection = (probabilities * target).sum()
        sums = probabilities.sum() + target.sum()
        j = intersection / (sums - intersection + eps)

        # this implements the support for looking just into the RoI
        h = torch.nn.functional.binary_cross_entropy_with_logits(
            input_, target, reduction="mean"
        )
        return (self.alpha * h) + ((1 - self.alpha) * (1 - j))


class MultiLayerBCELogitsLossWeightedPerBatch(BCEWithLogitsLossWeightedPerBatch):
    """Weighted Binary Cross-Entropy Loss for multi-layered inputs.

    This loss can be used in networks that produce more than one output that
    has to match output targets.  For example, architectures such as
    as :py:class:`.hed.HED` or :py:class:`.lwnet.LittleWNet` require this
    feature.

    It follows the inherited super class applying on-the-fly `pos_weight`
    updates per batch.
    """

    def __init__(self):
        super().__init__()

    def forward(self, input_: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        input_
            Logits produced by the model to be evaluated, with the shape ``[n,
            c]`` (classification), or ``[n, c, h, w]`` (segmentation).
        target
            Ground-truth information with the shape  ``[n, c]``
            (classification), or ``[n, c, h, w]`` (segmentation), containing
            zeroes and ones.

        Returns
        -------
            The average loss for all input data.
        """

        fwd = super().forward
        return torch.cat([fwd(i, target).unsqueeze(0) for i in input_]).mean()


class MultiLayerSoftJaccardAndBCELogitsLoss(SoftJaccardAndBCEWithLogitsLoss):
    """Implement Equation 3 in :cite:p:`iglovikov_ternausnetv2_2018` for the multi-output networks.

    This loss can be used in networks that produce more than one output that
    has to match output targets.  For example, architectures such as
    as :py:class:`.hed.HED` or :py:class:`.lwnet.LittleWNet` require this
    feature.

    Parameters
    ----------
    alpha : float
        Determines the weighting of SoftJaccard and BCE. Default: ``0.7``.
    """

    def __init__(self, alpha: float = 0.7):
        super().__init__(alpha=alpha)

    def forward(self, input_: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Parameters
        ----------
        input_
            Logits produced by the model to be evaluated, with the shape ``[n,
            c]`` (classification), or ``[n, c, h, w]`` (segmentation).
        target
            Ground-truth information with the shape  ``[n, c]``
            (classification), or ``[n, c, h, w]`` (segmentation), containing
            zeroes and ones.

        Returns
        -------
            The average loss for all input data.
        """

        fwd = super().forward
        return torch.cat([fwd(i, target).unsqueeze(0) for i in input_]).mean()


class MOONBCEWithLogitsLoss(torch.nn.Module):
    """Calculates the weighted binary cross entropy loss based on :cite:p:`guler_refining_2024`.

    This loss implements the domain-adapted multitask loss function in Equation
    (2) on :cite:p:`guler_refining_2024`.  The vector of input weights must be calculated from
    the input dataset in advance, and set during initialization, or later,
    before the loss can be fully used.

    Parameters
    ----------
    weights
        The positive weight of each class in the dataset given as input as a
        ``[2, C]`` tensor, with :math:`w_i^-` at position 0, and :math:`w_i^+`
        at position 1, as defined in Equation (1) of :cite:p:`guler_refining_2024`.
    """

    def __init__(self, weights: torch.Tensor | None = None):
        super().__init__()
        self.weights = weights

    @classmethod
    def _get_weight_from_targets(cls, targets: torch.Tensor) -> torch.Tensor:
        r"""Compute the MOON weights from a set of targets as per Equation
        (1) in :cite:p:`guler_refining_2024`.

        Parameters
        ----------
        targets
            A :py:class:`torch.Tensor` containing the targets, in the format
            ``[n, C]`` where ``n`` is the number of samples and ``C`` the
            number of classes.

        Returns
        -------
            The weight of each class in the dataset given as input as a ``[2,
            C]`` tensor, with :math:`w_i^-` at position 0, and :math:`w_i^+` at
            position 1, as defined in Equation (1) of :cite:p:`guler_refining_2024`.
        """

        task_type = _task_type(targets)

        if task_type == "segmentation":
            # rearranges ``targets`` vector so the problem looks like a simpler
            # classification problem where each pixel is a "separate sample"
            targets = (
                targets.transpose(0, 2).transpose(1, 3).reshape(-1, targets.shape[1])
            )

        # at this point targets should be [n x C]
        s_plus = targets.sum(dim=0)
        s_minus = targets.shape[0] - s_plus
        w_minus = (s_plus / s_minus).clamp(min=0, max=1)
        w_plus = (s_minus / s_plus).clamp(min=0, max=1)

        return torch.vstack((w_minus, w_plus))

    @classmethod
    def _get_weight_from_dataloader(
        cls, dataloader: torch.utils.data.DataLoader
    ) -> torch.Tensor:
        r"""Compute the MOON weights of each class of a DataLoader as per Equation
        (1) in :cite:p:`guler_refining_2024`.

        Parameters
        ----------
        dataloader
            A DataLoader from which to compute the positive weights.  Entries must
            be a dictionary which must contain a ``target`` key.

        Returns
        -------
            The weight of each class in the dataset given as input as a ``[2,
            C]`` tensor, with :math:`w_i^-` at position 0, and :math:`w_i^+` at
            position 1, as defined in Equation (1) of :cite:p:`guler_refining_2024`.
        """

        if isinstance(dataloader.dataset, Dataset):
            # there is a faster way to access the targets!
            targets = dataloader.dataset.targets()
        else:
            targets = [batch["target"] for batch in dataloader]

        return cls._get_weight_from_targets(torch.vstack(targets))

    @classmethod
    def get_arguments_from_datamodule(
        cls, datamodule: ConcatDataModule
    ) -> tuple[dict[str, torch.Tensor], dict[str, torch.Tensor]]:
        r"""Compute the MOON weights for train and validation sets of a datamodule.

        This function inputs a :py:class:`.data.datamodule.ConcatDataModule`,
        and for both the training and validation sets, and for each class on
        the respective dataloader targets, computes negative and positive
        weights as such:

        .. math::

           \begin{align}
               w_i^+ &=
                   \begin{cases}
                       1 & \text{if } S^{-}_{i} > S^{+}_{i} \\
                       \frac{S^{-}_{i}}{S^{+}_{i}} & \text{otherwise}
                   \end{cases} &
               w_i^- &=
                   \begin{cases}
                       1 & \text{if } S^{+}_{i} > S^{-}_{i} \\
                       \frac{S^{+}_{i}}{S^{-}_{i}} & \text{otherwise}
                   \end{cases}
           \end{align}

        This weight vector is used during runtime to balance individual batch
        losses respecting individual class distributions.

        Parameters
        ----------
        datamodule
            The datamodule to probe for training and validation datasets.

        Returns
        -------
            A tuple containing the training and validation ``weight``
            arguments, wrapped in a dictionary. Each ``weight`` variable
            contains the weights of each class in the target dataset as a ``[2,
            C]`` tensor, with :math:`w_i^-` at position 0, and :math:`w_i^+` at
            position 1, as defined in Equation (1) of :cite:p:`guler_refining_2024`.
        """

        train_weights = cls._get_weight_from_dataloader(
            datamodule.unshuffled_train_dataloader()
        )
        logger.info(f"train: MOONBCEWithLogitsLoss(weight={train_weights})")

        if "validation" in datamodule.val_dataloader().keys():
            validation_weights = cls._get_weight_from_dataloader(
                datamodule.val_dataloader()["validation"]
            )
        else:
            logger.warning(
                "Datamodule does not contain a validation dataloader. "
                "The training dataloader will be used instead."
            )
            validation_weights = train_weights
        logger.info(f"validation: MOONBCEWithLogitsLoss(weight={validation_weights})")

        return (dict(weights=train_weights), dict(weights=validation_weights))

    def forward(self, input_: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        r"""Forward pass.

        This function inputs the output of the model and a set of binary
        targets (as a float tensor containing zeroes and ones), and implements
        Equation (2) from :cite:p:`guler_refining_2024`:

        .. math::

           \mathcal J = -\sum_{i=1}^M w_i^{t_i} \bigl[t_i\log f_i(x) +
           (1-t_i)\log (1-f_i(x)) \bigr]

        Parameters
        ----------
        input_
            Logits produced by the model to be evaluated, with the shape ``[n,
            c]`` (classification), or ``[n, c, h, w]`` (segmentation).
        target
            Ground-truth information with the shape  ``[n, c]``
            (classification), or ``[n, c, h, w]`` (segmentation), containing
            zeroes and ones.

        Returns
        -------
            The result of Equation (2) from :cite:p:`guler_refining_2024`.

        Raises
        ------
        AssertionError
            In case the weights have not be initialized by calling
            :py:meth:`get_arguments_from_datamodule`.
        """

        assert self.weights is not None, (
            f"Weights are not initialized. Call "
            f"{self.__class__.__name__}.get_arguments_from_datamodule() to sort this."
        )

        if len(input_.shape) > 2:  # segmentation
            input_ = input_.transpose(0, 2).transpose(1, 3).reshape(-1, input_.shape[1])
            target = target.transpose(0, 2).transpose(1, 3).reshape(-1, target.shape[1])

        weights = ((1.0 - target) * self.weights[0]) + (target * self.weights[1])

        individual_losses = torch.nn.functional.binary_cross_entropy_with_logits(
            input_, target, reduction="none"
        )

        return (individual_losses * weights).mean()

    def to(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Self:
        """Move loss parameters to specified device.

        Refer to the method :py:meth:`torch.nn.Module.to` for details.

        Parameters
        ----------
        *args
            Parameter forwarded to the underlying implementations.
        **kwargs
            Parameter forwarded to the underlying implementations.

        Returns
        -------
            Self.
        """

        if self.weights is None:
            logger.warning(
                f"Weights are not initialized. Call {self.__class__.__name__}."
                f"get_arguments_from_datamodule() to sort this."
            )

            return self

        self.weights = self.weights.to(*args, **kwargs)

        return self



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Utilities to configure logging for command-line scripts."""

import logging


def setup_cli_logger() -> logging.Logger:
    """Set up the logger to be used for all CLI apps.

    Returns
    -------
        The logger to be used by CLI apps.
    """

    import colorlog
    from clapper.logging import setup

    logger = setup(
        __name__.split(".")[0],
        format="%(levelname)s: %(message)s",
        formatter=colorlog.ColoredFormatter(
            fmt="%(log_color)s[%(levelname)s]%(reset)s %(message)s",
            log_colors={
                "DEBUG": "light_black",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red,bg_white",
            },
        ),
    )
    logger.setLevel(logging.ERROR)

    return logger



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
""":py:mod:`Montgomery/Shenzhen aggregated database <mednet.data.classify.montgomery_shenzhen>` (cross validation fold 3)."""

import importlib.resources

from mednet.data.classify.montgomery_shenzhen import DataModule

datamodule = DataModule(
    "fold-3",
    (
        importlib.resources.files("mednet.config.classify.data.montgomery")
        / "fold-3.json",
        importlib.resources.files("mednet.config.classify.data.shenzhen")
        / "fold-3.json",
    ),
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Tests for the aggregated Montgomery-Shenzhen-Indian dataset."""

import importlib

import pytest
from click.testing import CliRunner


@pytest.mark.parametrize(
    "name",
    [
        "default",
        "fold_0",
        "fold_1",
        "fold_2",
        "fold_3",
        "fold_4",
        "fold_5",
        "fold_6",
        "fold_7",
        "fold_8",
        "fold_9",
    ],
)
def test_split_consistency(name: str):
    montgomery = importlib.import_module(
        f".{name}", "mednet.config.classify.data.montgomery"
    ).datamodule

    shenzhen = importlib.import_module(
        f".{name}", "mednet.config.classify.data.shenzhen"
    ).datamodule

    indian = importlib.import_module(
        f".{name}", "mednet.config.classify.data.indian"
    ).datamodule

    combined = importlib.import_module(
        f".{name}",
        "mednet.config.classify.data.montgomery_shenzhen_indian",
    ).datamodule

    montgomery_loader = importlib.import_module(
        ".montgomery", "mednet.data.classify"
    ).RawDataLoader

    shenzhen_loader = importlib.import_module(
        ".shenzhen", "mednet.data.classify"
    ).RawDataLoader

    indian_loader = importlib.import_module(
        ".indian", "mednet.data.classify"
    ).RawDataLoader

    for split in ("train", "validation", "test"):
        assert montgomery.splits[split][0][0] == combined.splits[split][0][0]
        assert isinstance(montgomery.splits[split][0][1], montgomery_loader)
        assert isinstance(combined.splits[split][0][1], montgomery_loader)

        assert shenzhen.splits[split][0][0] == combined.splits[split][1][0]
        assert isinstance(shenzhen.splits[split][0][1], shenzhen_loader)
        assert isinstance(combined.splits[split][1][1], shenzhen_loader)

        assert indian.splits[split][0][0] == combined.splits[split][2][0]
        assert isinstance(indian.splits[split][0][1], indian_loader)
        assert isinstance(combined.splits[split][2][1], indian_loader)


@pytest.mark.slow
@pytest.mark.skip_if_rc_var_not_set("datadir.montgomery")
@pytest.mark.skip_if_rc_var_not_set("datadir.indian")
@pytest.mark.skip_if_rc_var_not_set("datadir.shenzhen")
def test_database_check():
    from mednet.scripts.database import check

    runner = CliRunner()
    result = runner.invoke(check, ["montgomery-shenzhen-indian"])
    assert (
        result.exit_code == 0
    ), f"Exit code {result.exit_code} != 0 -- Output:\n{result.output}"



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Tests for cxr8 dataset."""

import importlib

import pytest
from click.testing import CliRunner


def id_function(val):
    if isinstance(val, dict):
        return str(val)
    return repr(val)


@pytest.mark.parametrize(
    "split,lengths",
    [
        ("default", dict(train=78484, validation=11212, test=22424)),
    ],
    ids=id_function,  # just changes how pytest prints it
)
def test_protocol_consistency(
    database_checkers,
    split: str,
    lengths: dict[str, int],
):
    from mednet.data.split import make_split

    database_checkers.check_split(
        make_split("mednet.config.segment.data.cxr8", f"{split}.json.bz2"),
        lengths=lengths,
        prefixes=[],
        possible_labels=[],
    )


@pytest.mark.skip_if_rc_var_not_set("datadir.cxr8")
def test_database_check():
    from mednet.scripts.database import check

    runner = CliRunner()
    result = runner.invoke(check, ["--limit=10", "cxr8"])
    assert (
        result.exit_code == 0
    ), f"Exit code {result.exit_code} != 0 -- Output:\n{result.output}"


@pytest.mark.skip_if_rc_var_not_set("datadir.cxr8")
@pytest.mark.parametrize(
    "dataset",
    [
        "train",
        "test",
    ],
)
@pytest.mark.parametrize(
    "name",
    [
        "default",
    ],
)
def test_loading(database_checkers, name: str, dataset: str):
    datamodule = importlib.import_module(
        f".{name}",
        "mednet.config.segment.data.cxr8",
    ).datamodule

    datamodule.model_transforms = []  # should be done before setup()
    datamodule.setup("predict")  # sets up all datasets

    loader = datamodule.predict_dataloader()[dataset]

    limit = 3  # limit load checking
    for batch in loader:
        if limit == 0:
            break
        database_checkers.check_loaded_batch(
            batch,
            batch_size=1,
            color_planes=3,
            prefixes=[],
            possible_labels=[],
            expected_num_labels=1,
            expected_meta_size=3,
        )
        limit -= 1


@pytest.mark.skip_if_rc_var_not_set("datadir.cxr8")
def test_raw_transforms_image_quality(database_checkers, datadir):
    datamodule = importlib.import_module(
        ".default",
        "mednet.config.segment.data.cxr8",
    ).datamodule

    datamodule.model_transforms = []
    datamodule.setup("predict")

    reference_histogram_file = (
        datadir / "histograms/raw_data/histograms_cxr8_default.json"
    )

    # Are we working with the original or pre-processed data set?
    # -> We retrieve the first sample and check its shape, if it is of the
    # expected size (1024x1024 pixels).  If not, we are working with a
    # preprocessed version of the database.  In this case, we check the
    # reference histogram against another file.
    is_original = (
        datamodule.unshuffled_train_dataloader().dataset[0]["image"].shape[-1] == 1024
    )
    if not is_original:
        reference_histogram_file = (
            reference_histogram_file.parent
            / "histograms_cxr8_preprocessed_default.json"
        )

    database_checkers.check_image_quality(datamodule, reference_histogram_file)


@pytest.mark.skip_if_rc_var_not_set("datadir.cxr8")
@pytest.mark.parametrize(
    "model_name",
    ["lwnet"],
)
def test_model_transforms_image_quality(database_checkers, datadir, model_name):
    reference_histogram_file = (
        datadir / f"histograms/models/histograms_{model_name}_cxr8_default.json"
    )

    datamodule = importlib.import_module(
        ".default",
        "mednet.config.segment.data.cxr8",
    ).datamodule

    model = importlib.import_module(
        f".{model_name}",
        "mednet.config.segment.models",
    ).model

    datamodule.model_transforms = model.model_transforms
    datamodule.setup("predict")

    database_checkers.check_image_quality(
        datamodule,
        reference_histogram_file,
        compare_type="statistical",
        pearson_coeff_threshold=0.005,
    )
